import json
import logging
from collections import defaultdict
from datetime import timedelta, timezone
from typing import IO, Dict, NamedTuple, Optional, Union, List
import warnings

from dateutil.parser import parse

from kloppy.domain import (
    BallState,
    DatasetFlag,
    Ground,
    Metadata,
    Orientation,
    Period,
    Player,
    PlayerData,
    Point,
    Point3D,
    PositionType,
    Provider,
    Team,
    TrackingDataset,
    AttackingDirection,
    Frame,
)

from kloppy.domain.services.frame_factory import create_frame
from kloppy.domain.services import attacking_direction_from_frame
from kloppy.exceptions import DeserializationError
from kloppy.infra.serializers.tracking.deserializer import (
    TrackingDataDeserializer,
)
from kloppy.utils import performance_logging

logger = logging.getLogger(__name__)


position_types_mapping: Dict[str, PositionType] = {
    "CB": PositionType.CenterBack,
    "LCB": PositionType.LeftCenterBack,
    "RCB": PositionType.RightCenterBack,
    "LB": PositionType.LeftBack,
    "RB": PositionType.RightBack,
    "DM": PositionType.DefensiveMidfield,
    "CM": PositionType.CenterMidfield,
    "LW": PositionType.LeftWing,
    "RW": PositionType.RightWing,
    "D": PositionType.Defender,
    "CF": PositionType.Striker,
    "M": PositionType.Midfielder,
    "AM": PositionType.AttackingMidfield,
    "GK": PositionType.Goalkeeper,
    "F": PositionType.Attacker,
}


class GameEventType:
    """
    GameEventType

    Type of game event.

    Attributes:
        FIRST_KICK_OFF: First half kick-off.
        SECOND_KICK_OFF: Second half kick-off.
        THIRD_KICK_OFF: Third half kick-off (if applicable).
        FOURTH_KICK_OFF: Fourth half kick-off (if applicable).
        PERIOD_END: End of half.
        STAYS_IN_PLAY: Ball hits post, bar, or corner flag and stays in play.
        PLAYER_OFF: Player off.
        PLAYER_ON: Player on.
        ON_THE_BALL: On-the-ball event.
        BALL_OUT: Ball out-of-play.
        SUBSTITUTION: Substitution event.
        VIDEO_MISSING: Video data missing.
    """

    FIRST_KICK_OFF = "FIRSTKICKOFF"
    SECOND_KICK_OFF = "SECONDKICKOFF"
    THIRD_KICK_OFF = "THIRDKICKOFF"
    FOURTH_KICK_OFF = "FOURTHKICKOFF"
    PERIOD_END = "END"
    STAYS_IN_PLAY = "G"
    PLAYER_OFF = "OFF"
    PLAYER_ON = "ON"
    ON_THE_BALL = "OTB"
    BALL_OUT = "OUT"
    SUBSTITUTION = "SUB"
    VIDEO_MISSING = "VID"


class PFF_TrackingInputs(NamedTuple):
    meta_data: IO[bytes]
    roster_meta_data: IO[bytes]
    raw_data: IO[bytes]


class PFF_TrackingDeserializer(TrackingDataDeserializer[PFF_TrackingInputs]):
    def __init__(
        self,
        limit: Optional[int] = None,
        sample_rate: Optional[float] = None,
        coordinate_system: Optional[Union[str, Provider]] = None,
        only_alive: Optional[bool] = False,
    ):
        super().__init__(limit, sample_rate, coordinate_system)
        self.only_alive = only_alive

        self._ball_owning_team = None
        self._ball_state = BallState.DEAD

    @property
    def provider(self) -> Provider:
        return Provider.PFF

    @classmethod
    def _get_frame_data(
        cls,
        players,
        periods,
        ball_owning_team,
        ball_state,
        frame,
    ):
        """Gets a Frame"""

        # Get Frame information
        frame_period = frame["period"]
        frame_id = frame["frameNum"]
        frame_timestamp = timedelta(seconds=frame["periodElapsedTime"])

        # Ball coordinates
        ball_smoothed = frame.get("ballsSmoothed")
        if ball_smoothed:
            ball_x = ball_smoothed.get("x")
            ball_y = ball_smoothed.get("y")
            ball_z = ball_smoothed.get("z")

            if ball_x is None or ball_y is None:
                ball_coordinates = None
            else:
                ball_coordinates = Point3D(
                    x=float(ball_x),
                    y=float(ball_y),
                    z=float(ball_z) if ball_z is not None else None,
                )

        else:
            ball_coordinates = None

        # Player coordinates
        players_data = {}

        # Helper function to map players
        def map_players(players_smoothed, team):
            if players_smoothed is None:
                return

            for player_info in players_smoothed:
                jersey_num = int(player_info.get("jerseyNum"))
                player = next(
                    (
                        p
                        for p in players[team].values()
                        if p.jersey_no == jersey_num
                    ),
                    None,
                )

                if player:
                    player_x = player_info.get("x")
                    player_y = player_info.get("y")

                    if player_x is not None and player_y is not None:
                        player_data = PlayerData(
                            coordinates=Point(player_x, player_y)
                        )
                        players_data[player] = player_data

        # Process home and away players
        map_players(frame.get("homePlayersSmoothed"), "HOME")
        map_players(frame.get("awayPlayersSmoothed"), "AWAY")

        return create_frame(
            frame_id=frame_id,
            timestamp=frame_timestamp,
            ball_coordinates=ball_coordinates,
            players_data=players_data,
            period=periods[frame_period],
            ball_state=ball_state,
            ball_owning_team=ball_owning_team,
            other_data={},
        )

    @classmethod
    def __get_periods(cls, tracking, frame_rate):
        """Gets the Periods contained in the tracking data"""
        periods = {}
        frames_by_period = defaultdict(list)

        for raw_frame in tracking:
            frame = json.loads(raw_frame)
            if frame["period"] is not None:
                frames_by_period[frame["period"]].append(frame)

        for period, frames in frames_by_period.items():
            periods[period] = Period(
                id=period,
                start_timestamp=timedelta(
                    seconds=frames[0]["frameNum"] / frame_rate
                ),
                end_timestamp=timedelta(
                    seconds=frames[-1]["frameNum"] / frame_rate
                ),
            )

        return periods

    def deserialize(self, inputs: PFF_TrackingInputs) -> TrackingDataset:
        # Load datasets
        metadata = json.load(inputs.meta_data)[0]
        roster_meta_data = json.load(inputs.roster_meta_data)
        raw_data = inputs.raw_data.readlines()

        # Obtain game_id from metadata
        game_id = int(metadata["id"])

        if not metadata:
            raise DeserializationError(
                "The game_id of this game is not contained within the provided meta data file"
            )

        # Get metadata variables
        home_team = metadata["homeTeam"]
        away_team = metadata["awayTeam"]
        stadium = metadata["stadium"]
        game_week = metadata["week"]

        # Obtain frame rate
        frame_rate = metadata["fps"]

        home_team_id = home_team["id"]
        away_team_id = away_team["id"]

        with performance_logging("Loading metadata", logger=logger):
            periods = self.__get_periods(raw_data, frame_rate)

            pitch_size_width = stadium["pitches"][0]["width"]
            pitch_size_length = stadium["pitches"][0]["length"]

            transformer = self.get_transformer(
                pitch_length=pitch_size_length, pitch_width=pitch_size_width
            )

            date = metadata.get("date")

            if date:
                date = parse(date).replace(tzinfo=timezone.utc)

            players = {"HOME": {}, "AWAY": {}}

            # Create Team objects for home and away sides
            home_team = Team(
                team_id=home_team_id,
                name=home_team["name"],
                ground=Ground.HOME,
            )
            away_team = Team(
                team_id=away_team_id,
                name=away_team["name"],
                ground=Ground.AWAY,
            )
            teams = [home_team, away_team]

            for player in roster_meta_data:

                team_id = player["team"]["id"]
                player_col = player["player"]

                player_id = player_col["id"]
                player_name = player_col["nickname"]
                shirt_number = int(player["shirtNumber"])
                player_position = player["positionGroupType"]

                if team_id == home_team_id:
                    team_string = "HOME"
                    team = home_team
                elif team_id == away_team_id:
                    team_string = "AWAY"
                    team = away_team
                else:
                    raise DeserializationError(f"Unknown team_id: {team_id}")

                # Create Player object
                players[team_string][player_id] = Player(
                    player_id=player_id,
                    team=team,
                    jersey_no=shirt_number,
                    name=player_name,
                    starting_position=position_types_mapping.get(
                        player_position, PositionType.Unknown
                    ),
                    starting=None,
                )

            home_team.players = list(players["HOME"].values())
            away_team.players = list(players["AWAY"].values())

        with performance_logging("Loading data", logger=logger):

            def _iter():
                sample = 1.0 / self.sample_rate

                n = 0

                for raw_frame in raw_data:
                    frame = json.loads(raw_frame)
                    # Identify Period
                    frame_period = frame.get("period")

                    # Find ball owning team
                    game_event = frame.get("game_event")

                    if game_event:
                        game_event_type = game_event.get("game_event_type")

                        if game_event_type in (
                            GameEventType.BALL_OUT,
                            GameEventType.PERIOD_END,
                        ):
                            self._ball_state = BallState.DEAD
                        elif game_event_type in (
                            GameEventType.FIRST_KICK_OFF,
                            GameEventType.SECOND_KICK_OFF,
                            GameEventType.THIRD_KICK_OFF,
                            GameEventType.FOURTH_KICK_OFF,
                            GameEventType.ON_THE_BALL,
                        ):
                            self._ball_state = BallState.ALIVE
                        else:
                            # for other events leave ball state as is
                            pass

                        if game_event.get("home_ball") is not None:
                            self._ball_owning_team = (
                                home_team
                                if game_event["home_ball"]
                                else away_team
                            )

                    if self.only_alive and self._ball_state == BallState.DEAD:
                        continue

                    if frame_period is not None and n % sample == 0:
                        yield frame, frame_period

                    n += 1

        frames, et_frames = [], []
        n_frames = 0

        for _frame, _frame_period in _iter():
            # Create and transform Frame object
            frame = transformer.transform_frame(
                self._get_frame_data(
                    players,
                    periods,
                    self._ball_owning_team,
                    self._ball_state,
                    _frame,
                )
            )

            # if Regular Time
            if _frame_period in {1, 2}:
                frames.append(frame)

            # else if extra time
            elif _frame_period in {3, 4}:
                et_frames.append(frame)

            n_frames += 1

            if self.limit and n_frames >= self.limit:
                break

        first_frame_p1 = next(
            frame for frame in frames if frame.period.id == 1
        )
        is_home_team_left = (
            True
            if attacking_direction_from_frame(first_frame_p1)
            == AttackingDirection.LTR
            else False
        )

        if et_frames:
            first_frame_p3 = next(
                frame for frame in et_frames if frame.period.id == 3
            )
            is_home_team_extra_time_left = (
                True
                if attacking_direction_from_frame(first_frame_p3)
                == AttackingDirection.LTR
                else False
            )
            # If first period and third period attacking direction for home team is inconsistent, flip the direction of the extra time frames
            if is_home_team_left != is_home_team_extra_time_left:
                for et_frame in et_frames:
                    # Loop through each PlayerData in the players_data dictionary
                    for _, player_data in et_frame.players_data.items():
                        if (
                            player_data.coordinates
                            and player_data.coordinates.x is not None
                            and player_data.coordinates.y is not None
                        ):
                            # Create a new Point with multiplied coordinates for each player
                            player_data.coordinates = Point(
                                -player_data.coordinates.x,
                                -player_data.coordinates.y,
                            )

                    # Multiply the x and y coordinates of the ball by -1
                    if (
                        et_frame.ball_coordinates
                        and et_frame.ball_coordinates.x is not None
                        and et_frame.ball_coordinates.y is not None
                    ):
                        et_frame.ball_coordinates = Point3D(
                            -et_frame.ball_coordinates.x,
                            -et_frame.ball_coordinates.y,
                            et_frame.ball_coordinates.z,
                        )

        orientation = (
            Orientation.HOME_AWAY
            if is_home_team_left
            else Orientation.AWAY_HOME
        )

        frames.extend(et_frames)

        metadata = Metadata(
            teams=teams,
            periods=sorted(periods.values(), key=lambda p: p.id),
            pitch_dimensions=transformer.get_to_coordinate_system().pitch_dimensions,
            frame_rate=frame_rate,
            orientation=orientation,
            provider=Provider.PFF,
            flags=DatasetFlag.BALL_OWNING_TEAM | DatasetFlag.BALL_STATE,
            coordinate_system=transformer.get_to_coordinate_system(),
            date=date,
            game_id=game_id,
            game_week=game_week,
        )

        return TrackingDataset(
            records=frames,
            metadata=metadata,
        )



from datetime import datetime, timedelta, timezone
from pathlib import Path

import pytest

from kloppy import wyscout
from kloppy.domain import (
    BodyPart,
    BodyPartQualifier,
    CardQualifier,
    CardType,
    DatasetType,
    DuelQualifier,
    DuelType,
    EventDataset,
    EventType,
    FormationType,
    GoalkeeperActionType,
    GoalkeeperQualifier,
    Orientation,
    PassQualifier,
    PassResult,
    PassType,
    Point,
    PositionType,
    FormationType,
    SetPieceQualifier,
    SetPieceType,
    ShotResult,
    Time,
    Point3D,
)


@pytest.fixture(scope="session")
def event_v2_data(base_dir: Path) -> Path:
    return base_dir / "files" / "wyscout_events_v2.json"


@pytest.fixture(scope="session")
def event_v3_data(base_dir: Path) -> Path:
    return base_dir / "files" / "wyscout_events_v3.json"


def test_correct_auto_recognize_deserialization(
    event_v2_data: Path, event_v3_data: Path
):
    dataset = wyscout.load(event_data=event_v2_data, coordinates="wyscout")
    assert dataset.records[2].coordinates == Point(29.0, 6.0)
    dataset = wyscout.load(event_data=event_v3_data, coordinates="wyscout")
    assert dataset.records[2].coordinates == Point(32.0, 56.0)


class TestWyscoutV2:
    """Tests related to deserialization of Wyscout V2 data."""

    @pytest.fixture(scope="class")
    def dataset(self, event_v2_data) -> EventDataset:
        """Load Wyscout V2 event dataset"""
        dataset = wyscout.load(
            event_data=event_v2_data,
            coordinates="wyscout",
            data_version="V2",
        )
        assert dataset.dataset_type == DatasetType.EVENT
        assert (
            dataset.metadata.orientation == Orientation.ACTION_EXECUTING_TEAM
        )
        return dataset

    def test_metadata(self, dataset: EventDataset):
        assert dataset.metadata.periods[0].id == 1
        assert dataset.metadata.periods[0].start_timestamp == timedelta(
            seconds=0
        )
        assert dataset.metadata.periods[0].end_timestamp == timedelta(
            seconds=2863.708369
        )
        assert dataset.metadata.periods[1].id == 2
        assert dataset.metadata.periods[1].start_timestamp == timedelta(
            seconds=2863.708369
        )
        assert dataset.metadata.periods[1].end_timestamp == timedelta(
            seconds=2863.708369
        ) + timedelta(seconds=2999.70982)

        game_id = dataset.metadata.game_id
        if game_id:
            assert isinstance(game_id, str)
            assert game_id == "2499773"

    def test_timestamps(self, dataset: EventDataset):
        kickoff_p1 = dataset.get_event_by_id("190078343")
        assert kickoff_p1.timestamp == timedelta(seconds=2.643377)
        kickoff_p2 = dataset.get_event_by_id("190079822")
        assert kickoff_p2.timestamp == timedelta(seconds=0)

    def test_shot_event(self, dataset: EventDataset):
        shot_event = dataset.get_event_by_id("190079151")
        assert (
            shot_event.get_qualifier_value(BodyPartQualifier)
            == BodyPart.RIGHT_FOOT
        )

    def test_miscontrol_event(self, dataset: EventDataset):
        miscontrol_event = dataset.get_event_by_id("190078351")
        assert miscontrol_event.event_type == EventType.MISCONTROL

    def test_interception_event(self, dataset: EventDataset):
        # A touch or duel with "interception" tag should be converted to an interception event
        interception_event = dataset.get_event_by_id("190079090")
        assert interception_event.event_type == EventType.INTERCEPTION
        # Other events with "interception" tag should be split in two events
        clearance_event = dataset.get_event_by_id("190079171")
        assert clearance_event.event_type == EventType.CLEARANCE
        interception_event = dataset.get_event_by_id("interception-190079171")
        assert interception_event.event_type == EventType.INTERCEPTION

    def test_duel_event(self, dataset: EventDataset):
        ground_duel_event = dataset.get_event_by_id("190078379")
        assert ground_duel_event.event_type == EventType.DUEL
        assert (
            ground_duel_event.get_qualifier_value(DuelQualifier)
            == DuelType.GROUND
        )
        aerial_duel_event = dataset.get_event_by_id("190078381")
        assert aerial_duel_event.event_type == EventType.DUEL
        assert (
            aerial_duel_event.get_qualifier_values(DuelQualifier)[1]
            == DuelType.AERIAL
        )
        sliding_tackle_duel_event = dataset.get_event_by_id("190079260")
        assert sliding_tackle_duel_event.event_type == EventType.DUEL
        assert (
            sliding_tackle_duel_event.get_qualifier_values(DuelQualifier)[2]
            == DuelType.SLIDING_TACKLE
        )

    def test_goalkeeper_event(self, dataset: EventDataset):
        goalkeeper_event = dataset.get_event_by_id("190079010")
        assert goalkeeper_event.event_type == EventType.GOALKEEPER
        assert (
            goalkeeper_event.get_qualifier_value(GoalkeeperQualifier)
            == GoalkeeperActionType.SAVE
        )

    def test_foul_committed_event(self, dataset: EventDataset):
        foul_committed_event = dataset.get_event_by_id("190079289")
        assert foul_committed_event.event_type == EventType.FOUL_COMMITTED
        assert (
            foul_committed_event.get_qualifier_value(CardQualifier)
            == CardType.FIRST_YELLOW
        )
        card_event = dataset.get_event_by_id("card-190079289")
        assert card_event.event_type == EventType.CARD

    def test_correct_normalized_deserialization(self, event_v2_data: Path):
        dataset = wyscout.load(event_data=event_v2_data, data_version="V2")
        assert dataset.records[2].coordinates == Point(
            0.2981354967264447, 0.06427244582043344
        )


class TestWyscoutV3:
    """Tests related to deserialization of Wyscout V3 data."""

    @pytest.fixture(scope="class")
    def dataset(self, event_v3_data: Path) -> EventDataset:
        """Load Wyscout V3 event dataset"""
        dataset = wyscout.load(
            event_data=event_v3_data,
            coordinates="wyscout",
            data_version="V3",
        )
        assert dataset.dataset_type == DatasetType.EVENT
        assert (
            dataset.metadata.orientation == Orientation.ACTION_EXECUTING_TEAM
        )
        return dataset

    def test_metadata(self, dataset: EventDataset):
        assert dataset.metadata.periods[0].id == 1
        assert dataset.metadata.periods[0].start_timestamp == timedelta(
            seconds=0
        )
        assert dataset.metadata.periods[0].end_timestamp == timedelta(
            minutes=45, seconds=5
        )
        assert dataset.metadata.periods[1].id == 2
        assert dataset.metadata.periods[1].start_timestamp == timedelta(
            minutes=45, seconds=5
        )
        assert dataset.metadata.periods[1].end_timestamp == timedelta(
            minutes=45, seconds=5
        ) + timedelta(minutes=46, seconds=58)

        assert (
            dataset.metadata.teams[0].starting_formation
            == FormationType.THREE_FOUR_TWO_ONE
        )

        formation_time_change = Time(
            dataset.metadata.periods[1], timedelta(seconds=3)
        )
        assert (
            dataset.metadata.teams[1].formations.items[formation_time_change]
            == FormationType.FOUR_THREE_ONE_TWO
        )

        second_period_end_time = Time(
            period=dataset.metadata.periods[1],
            timestamp=timedelta(seconds=2818),
        )
        assert (
            dataset.metadata.teams[1]
            .formations.items.keys()[0]
            .period.end_time
            == second_period_end_time
        )

        cr7 = dataset.metadata.teams[0].get_player_by_id("3322")

        assert cr7.full_name == "Cristiano Ronaldo dos Santos Aveiro"
        assert cr7.starting is True
        assert cr7.positions.last() == PositionType.Striker
        assert cr7.jersey_no == 7

    def test_enriched_metadata(self, dataset: EventDataset):
        date = dataset.metadata.date
        if date:
            assert isinstance(date, datetime)
            assert date == datetime(2020, 9, 20, 18, 45, tzinfo=timezone.utc)

        game_week = dataset.metadata.game_week
        if game_week:
            assert isinstance(game_week, str)
            assert game_week == "1"

        game_id = dataset.metadata.game_id
        if game_id:
            assert isinstance(game_id, str)
            assert game_id == "5154199"

    def test_timestamps(self, dataset: EventDataset):
        kickoff_p1 = dataset.get_event_by_id(1927028854)
        assert kickoff_p1.timestamp == timedelta(minutes=0, seconds=3)
        assert kickoff_p1.time.period.id == 1
        kickoff_p2 = dataset.get_event_by_id(1927029460)
        assert kickoff_p2.timestamp == timedelta(minutes=0, seconds=0)
        assert kickoff_p2.time.period.id == 2

    def test_coordinates(self, dataset: EventDataset):
        assert dataset.records[2].coordinates == Point(32.0, 56.0)

    def test_normalized_deserialization(self, event_v3_data: Path):
        dataset = wyscout.load(event_data=event_v3_data, data_version="V3")
        assert dataset.records[2].coordinates == Point(
            x=0.32643853442316295, y=0.5538235294117646
        )

    def test_pass_event(self, dataset: EventDataset):
        pass_event = dataset.get_event_by_id(1927028486)
        assert pass_event.event_type == EventType.PASS
        assert pass_event.coordinates == Point(x=22.0, y=91.0)
        assert pass_event.receiver_coordinates == Point(x=8.0, y=71.0)

        blocked_pass_event = dataset.get_event_by_id(1927029452)
        assert blocked_pass_event.result == PassResult.INCOMPLETE
        assert blocked_pass_event.coordinates == Point(x=96.0, y=85.0)
        assert blocked_pass_event.receiver_coordinates == Point(x=99.0, y=84.0)

    def test_goalkeeper_event(self, dataset: EventDataset):
        goalkeeper_event = dataset.get_event_by_id(1927029095)
        assert goalkeeper_event.event_type == EventType.GOALKEEPER
        assert (
            goalkeeper_event.get_qualifier_value(GoalkeeperQualifier)
            == GoalkeeperActionType.SAVE
        )

    def test_shot_assist_event(self, dataset: EventDataset):
        shot_assist_event = dataset.get_event_by_id(1927028561)
        assert shot_assist_event.event_type == EventType.PASS
        assert PassType.SHOT_ASSIST in shot_assist_event.get_qualifier_values(
            PassQualifier
        )

    def test_shot_event(self, dataset: EventDataset):
        # a blocked free kick shot
        blocked_shot_event = dataset.get_event_by_id(1927028534)
        assert blocked_shot_event.event_type == EventType.SHOT
        assert blocked_shot_event.result == ShotResult.BLOCKED
        assert blocked_shot_event.result_coordinates == Point(x=77.0, y=21.0)
        assert (
            blocked_shot_event.get_qualifier_value(SetPieceQualifier)
            == SetPieceType.FREE_KICK
        )
        # off target shot
        off_target_shot = dataset.get_event_by_id(1927028562)
        assert off_target_shot.event_type == EventType.SHOT
        assert off_target_shot.result == ShotResult.OFF_TARGET
        assert off_target_shot.result_coordinates == Point3D(
            x=100, y=40, z=3.5
        )
        # on target shot
        on_target_shot = dataset.get_event_by_id(1927028637)
        assert on_target_shot.event_type == EventType.SHOT
        assert on_target_shot.result == ShotResult.SAVED
        assert on_target_shot.result_coordinates == Point3D(x=100, y=45, z=1)

    def test_foul_committed_event(self, dataset: EventDataset):
        foul_committed_event = dataset.get_event_by_id(1927028873)
        assert foul_committed_event.event_type == EventType.FOUL_COMMITTED

    def test_duel_event(self, dataset: EventDataset):
        ground_duel_event = dataset.get_event_by_id(1927028474)
        assert ground_duel_event.event_type == EventType.DUEL
        assert (
            ground_duel_event.get_qualifier_value(DuelQualifier)
            == DuelType.GROUND
        )
        aerial_loose_ball_duel_event = dataset.get_event_by_id(1927028472)
        assert aerial_loose_ball_duel_event.event_type == EventType.DUEL
        assert (
            DuelType.LOOSE_BALL
            in aerial_loose_ball_duel_event.get_qualifier_values(DuelQualifier)
        )
        assert (
            DuelType.AERIAL
            in aerial_loose_ball_duel_event.get_qualifier_values(DuelQualifier)
        )
        sliding_tackle_duel_event = dataset.get_event_by_id(1927028828)
        assert sliding_tackle_duel_event.event_type == EventType.DUEL
        assert (
            DuelType.SLIDING_TACKLE
            in sliding_tackle_duel_event.get_qualifier_values(DuelQualifier)
        )

    def test_clearance_event(self, dataset: EventDataset):
        clearance_event = dataset.get_event_by_id(1927028482)
        assert clearance_event.event_type == EventType.CLEARANCE

    def test_interception_event(self, dataset: EventDataset):
        interception_event = dataset.get_event_by_id(1927028880)
        assert interception_event.event_type == EventType.INTERCEPTION

    def test_take_on_event(self, dataset: EventDataset):
        take_on_event = dataset.get_event_by_id(1927028870)
        assert take_on_event.event_type == EventType.TAKE_ON

    def test_carry_event(self, dataset: EventDataset):
        carry_event = dataset.get_event_by_id(1927028490)
        assert carry_event.event_type == EventType.CARRY
        assert carry_event.end_coordinates == Point(17.0, 4.0)

    def test_formation_change_event(self, dataset: EventDataset):
        assert (
            len(dataset.find_all("formation_change")) == 2
        )  # We shouldn't recognize the change to 4-4-1 as a formation change
        formation_change_event = dataset.get_event_by_id(
            "synthetic-3164-1927029462"
        )
        assert formation_change_event.event_type == EventType.FORMATION_CHANGE
        assert (
            formation_change_event.formation_type
            == FormationType.FOUR_THREE_ONE_TWO
        )

    def test_kick_off_qualifier(self, dataset: EventDataset):
        pass_event_kick_off_first_half = dataset.get_event_by_id(1927028854)
        pass_event_kick_off_second_half = dataset.get_event_by_id(1927029460)
        pass_event_kick_off_after_goal = dataset.get_event_by_id(1927030641)
        assert pass_event_kick_off_first_half.event_type == EventType.PASS
        assert pass_event_kick_off_second_half.event_type == EventType.PASS
        assert pass_event_kick_off_after_goal.event_type == EventType.PASS
        assert (
            SetPieceType.KICK_OFF
            in pass_event_kick_off_first_half.get_qualifier_values(
                SetPieceQualifier
            )
        )
        assert (
            SetPieceType.KICK_OFF
            in pass_event_kick_off_second_half.get_qualifier_values(
                SetPieceQualifier
            )
        )
        assert (
            SetPieceType.KICK_OFF
            in pass_event_kick_off_after_goal.get_qualifier_values(
                SetPieceQualifier
            )
        )

    def test_through_ball_qualifier(self, dataset: EventDataset):
        pass_event = dataset.get_event_by_id(1927028612)
        assert pass_event.event_type == EventType.PASS
        assert PassType.THROUGH_BALL in pass_event.get_qualifier_values(
            PassQualifier
        )

    def test_high_pass_qualifier(self, dataset: EventDataset):
        pass_event = dataset.get_event_by_id(1927028860)
        assert pass_event.event_type == EventType.PASS
        assert PassType.HIGH_PASS in pass_event.get_qualifier_values(
            PassQualifier
        )



from datetime import datetime, timedelta, timezone
from pathlib import Path

import pytest

from kloppy.domain import (
    Provider,
    BallState,
    Orientation,
    Point,
    Point3D,
    DatasetType,
)

from kloppy import skillcorner


class TestSkillCornerTracking:
    @pytest.fixture
    def meta_data(self, base_dir) -> str:
        return base_dir / "files/skillcorner_match_data.json"

    @pytest.fixture
    def raw_data(self, base_dir) -> str:
        return base_dir / "files/skillcorner_structured_data.json"

    @pytest.fixture
    def meta_data_v3(self, base_dir) -> str:
        return base_dir / "files/skillcorner_meta_data.json"

    @pytest.fixture
    def raw_data_v3(self, base_dir) -> str:
        return base_dir / "files/skillcorner_v3_raw_data.jsonl"

    @pytest.fixture
    def raw_data_timestamp(self, base_dir) -> str:
        return base_dir / "files/skillcorner_structured_data_timestamp.json"

    def test_correct_deserialization_timestamp(
        self, raw_data_timestamp: Path, meta_data: Path
    ):
        skillcorner.load(
            meta_data=meta_data,
            raw_data=raw_data_timestamp,
            coordinates="skillcorner",
            include_empty_frames=True,
        )

    def test_correct_deserialization(self, raw_data: Path, meta_data: Path):
        dataset = skillcorner.load(
            meta_data=meta_data,
            raw_data=raw_data,
            coordinates="skillcorner",
            include_empty_frames=True,
        )

        assert dataset.metadata.provider == Provider.SKILLCORNER
        assert dataset.dataset_type == DatasetType.TRACKING
        assert len(dataset.records) == 55632
        assert len(dataset.metadata.periods) == 2
        assert dataset.metadata.orientation == Orientation.AWAY_HOME
        assert dataset.metadata.periods[0].id == 1
        assert dataset.metadata.periods[0].start_timestamp == timedelta(
            seconds=1411 / 10
        )
        assert dataset.metadata.periods[0].end_timestamp == timedelta(
            seconds=28944 / 10
        )
        assert dataset.metadata.periods[1].id == 2
        assert dataset.metadata.periods[1].start_timestamp == timedelta(
            seconds=39979 / 10
        )
        assert dataset.metadata.periods[1].end_timestamp == timedelta(
            seconds=68076 / 10
        )

        assert dataset.records[0].frame_id == 1411
        assert dataset.records[0].timestamp == timedelta(seconds=0)
        assert dataset.records[27534].frame_id == 39979
        assert dataset.records[27534].timestamp == timedelta(seconds=0)

        # make sure skillcorner ID is used as player ID
        assert dataset.metadata.teams[0].players[0].player_id == "10247"

        # make sure data is loaded correctly
        home_player = dataset.metadata.teams[0].players[2]
        assert dataset.records[112].players_data[
            home_player
        ].coordinates == Point(x=33.8697315398, y=-9.55742259253)

        away_player = dataset.metadata.teams[1].players[9]
        assert dataset.records[112].players_data[
            away_player
        ].coordinates == Point(x=25.9863082795, y=27.3013598578)

        assert dataset.records[113].ball_coordinates == Point3D(
            x=30.5914728131, y=35.3622277834, z=2.24371228757
        )

        # check that missing ball-z_coordinate is identified as None
        assert dataset.records[150].ball_coordinates == Point3D(
            x=11.6568802848, y=24.7214038909, z=None
        )

        # check that 'ball_z' column is included in to_pandas dataframe
        # frame = _frame_to_pandas_row_converter(dataset.records[150])
        # assert "ball_z" in frame.keys()

        # make sure player data is only in the frame when the player is in view
        assert "home_1" not in [
            player.player_id
            for player in dataset.records[112].players_data.keys()
        ]

        assert "away_1" not in [
            player.player_id
            for player in dataset.records[112].players_data.keys()
        ]

        # are anonymous players loaded correctly?
        home_anon_75 = [
            player
            for player in dataset.records[197].players_data
            if player.player_id == "home_anon_75"
        ]
        assert home_anon_75 == [
            player
            for player in dataset.records[200].players_data
            if player.player_id == "home_anon_75"
        ]

        # is pitch dimension set correctly?
        pitch_dimensions = dataset.metadata.pitch_dimensions
        assert pitch_dimensions.x_dim.min == -52.5
        assert pitch_dimensions.x_dim.max == 52.5
        assert pitch_dimensions.y_dim.min == -34
        assert pitch_dimensions.y_dim.max == 34

        # Check enriched metadata
        date = dataset.metadata.date
        if date:
            assert isinstance(date, datetime)
            assert date == datetime(
                2019, 11, 9, 17, 30, 0, tzinfo=timezone.utc
            )

        game_id = dataset.metadata.game_id
        if game_id:
            assert isinstance(game_id, str)
            assert game_id == "2417"

        home_coach = dataset.metadata.home_coach
        if home_coach:
            assert isinstance(home_coach, str)
            assert home_coach == "Hans-Dieter Flick"

        away_coach = dataset.metadata.away_coach
        if away_coach:
            assert isinstance(away_coach, str)
            assert away_coach == "Lucien Favre"

    def test_correct_normalized_deserialization(
        self, meta_data: str, raw_data: str
    ):
        dataset = skillcorner.load(meta_data=meta_data, raw_data=raw_data)

        home_player = dataset.metadata.teams[0].players[2]
        assert dataset.records[0].players_data[
            home_player
        ].coordinates == Point(x=0.8225688718076191, y=0.6405503322430883)

    def test_skip_empty_frames(self, meta_data: str, raw_data: str):
        dataset = skillcorner.load(
            meta_data=meta_data, raw_data=raw_data, include_empty_frames=False
        )

        assert len(dataset.records) == 34783
        assert dataset.records[0].timestamp == timedelta(seconds=11.2)
        assert dataset.records[-1].ball_state == BallState.ALIVE

    def test_correct_deserialization_v3(
        self, raw_data_v3: Path, meta_data_v3: Path
    ):
        dataset = skillcorner.load(
            meta_data=meta_data_v3,
            raw_data=raw_data_v3,
            coordinates="skillcorner",
            include_empty_frames=True,
        )

        assert dataset.metadata.provider == Provider.SKILLCORNER
        assert dataset.dataset_type == DatasetType.TRACKING
        assert len(dataset.records) == 27
        assert len(dataset.metadata.periods) == 2
        assert dataset.metadata.periods[0].id == 1
        assert dataset.metadata.periods[0].start_timestamp == timedelta(
            seconds=1
        )
        assert dataset.metadata.periods[0].end_timestamp == timedelta(
            seconds=2, microseconds=200000
        )
        assert dataset.metadata.periods[1].id == 2
        assert dataset.metadata.periods[1].start_timestamp == timedelta(
            seconds=6097, microseconds=700000
        )
        assert dataset.metadata.periods[1].end_timestamp == timedelta(
            seconds=6099
        )

        assert dataset.records[0].frame_id == 10
        assert dataset.records[0].timestamp == timedelta(seconds=0)
        assert dataset.records[-1].frame_id == 60990
        assert dataset.records[-1].timestamp == timedelta(seconds=3256)

        home_team_gk = dataset.metadata.teams[0].get_player_by_id("133")
        assert home_team_gk.player_id == "133"
        assert dataset.records[10].players_data[
            home_team_gk
        ].coordinates == Point(x=40.46, y=-0.58)

        away_team_gk = dataset.metadata.teams[1].get_player_by_id("76")
        assert away_team_gk.player_id == "76"
        assert dataset.records[10].players_data[
            away_team_gk
        ].coordinates == Point(x=-41.97, y=-0.61)

        assert dataset.records[-1].ball_state == BallState.ALIVE
        assert dataset.records[-2].ball_state == BallState.DEAD



from typing import Iterable, Optional, Union

from kloppy.domain import TrackingDataset
from kloppy.infra.serializers.tracking.hawkeye import (
    HawkEyeDeserializer,
    HawkEyeInputs,
)
from kloppy.io import FileLike, expand_inputs


def load(
    ball_feeds: Union[FileLike, Iterable[FileLike]],
    player_centroid_feeds: Union[FileLike, Iterable[FileLike]],
    meta_data: Optional[FileLike] = None,
    pitch_width: Optional[float] = 68.0,
    pitch_length: Optional[float] = 105.0,
    sample_rate: Optional[float] = None,
    limit: Optional[int] = None,
    coordinates: Optional[str] = None,
    show_progress: Optional[bool] = False,
) -> TrackingDataset:
    """
    Load HawkEye tracking data.

    HawkEye splits up the data of a single game into multiple files, typically
    one file per minute. The `ball_feeds` and `player_centroid_feeds` arguments
    support various options on how to provide these files:

        - A file-like object to load the data for a single minute of the match.
        - An ordered list of file-like objects to load the data for multiple
          minutes of the match.
        - A directory containing the files to load. This requires that the ball
          feeds end with `.samples.ball` and the player centroid feeds end with
          `.samples.centroids`.


    Args:
        ball_feeds: The ball tracking data.
        player_centroid_feeds: The player centroid tracking data.
        meta_data: A json or xml file with metadata about the match. Metadata is optional.
        pitch_length: The length of the pitch (in meters). Ignored if metadata that includes pitch length is provided.
        pitch_width: The width of the pitch (in meters). Ignored if metadata that includes pitch width is provided.
        sample_rate: Sample the data at a specific rate.
        limit: Limit the number of frames to load to the first `limit` frames.
        coordinates: The coordinate system to use.
        show_progress: Show a progress bar while parsing the data.

    Returns:
        The parsed tracking data.

    Note:
        Pose tracking data is not yet supported.
    """
    ball_feeds = expand_inputs(ball_feeds, regex_filter="samples.ball")
    player_centroid_feeds = expand_inputs(
        player_centroid_feeds, regex_filter="samples.centroids"
    )

    deserializer = HawkEyeDeserializer(
        pitch_width=pitch_width,
        pitch_length=pitch_length,
        sample_rate=sample_rate,
        limit=limit,
        coordinate_system=coordinates,
    )
    return deserializer.deserialize(
        inputs=HawkEyeInputs(
            ball_feeds=ball_feeds,
            player_centroid_feeds=player_centroid_feeds,
            meta_data=meta_data,
            show_progress=show_progress,
        )
    )



import json
import logging
from dataclasses import replace
from datetime import timedelta
from typing import Dict, List, NamedTuple, IO, Optional

from kloppy.domain import (
    BodyPart,
    BodyPartQualifier,
    CardQualifier,
    CardType,
    CounterAttackQualifier,
    DuelResult,
    DuelQualifier,
    DuelType,
    EventDataset,
    EventType,
    GoalkeeperQualifier,
    GoalkeeperActionType,
    Ground,
    InterceptionResult,
    Metadata,
    Orientation,
    PassQualifier,
    PassResult,
    PassType,
    Period,
    Player,
    Point,
    Provider,
    Qualifier,
    SetPieceQualifier,
    SetPieceType,
    ShotResult,
    Team,
)
from kloppy.utils import performance_logging

from . import wyscout_events, wyscout_tags
from ..deserializer import EventDataDeserializer

logger = logging.getLogger(__name__)


INVALID_PLAYER = "0"


def _parse_team(raw_events, wyId: str, ground: Ground) -> Team:
    team = Team(
        team_id=wyId,
        name=raw_events["teams"][wyId]["team"]["officialName"],
        ground=ground,
    )
    team.players = [
        Player(
            player_id=str(player["player"]["wyId"]),
            team=team,
            jersey_no=None,
            first_name=player["player"]["firstName"],
            last_name=player["player"]["lastName"],
        )
        for player in raw_events["players"][wyId]
    ]
    return team


def _has_tag(raw_event, tag_id) -> bool:
    for tag in raw_event["tags"]:
        if tag["id"] == tag_id:
            return True
    return False


def _generic_qualifiers(raw_event: Dict) -> List[Qualifier]:
    qualifiers: List[Qualifier] = []

    if _has_tag(raw_event, wyscout_tags.COUNTER_ATTACK):
        qualifiers.append(CounterAttackQualifier(True))
    else:
        qualifiers.append(CounterAttackQualifier(False))

    return qualifiers


def _bodypart_qualifiers(raw_event: Dict) -> List[Qualifier]:
    qualifiers = []
    if _has_tag(raw_event, wyscout_tags.LEFT_FOOT):
        qualifiers.append(BodyPartQualifier(BodyPart.LEFT_FOOT))
    elif _has_tag(raw_event, wyscout_tags.RIGHT_FOOT):
        qualifiers.append(BodyPartQualifier(BodyPart.RIGHT_FOOT))
    elif _has_tag(raw_event, wyscout_tags.HEAD_BODY):
        qualifiers.append(BodyPartQualifier(BodyPart.HEAD_OTHER))
    return qualifiers


def _create_shot_result_coordinates(raw_event: Dict) -> Optional[Point]:
    """Estimate the shot end location from the Wyscout tags.

    Wyscout does not provide end-coordinates of shots. Instead shots on goal
    are tagged with a zone. This function maps each of these zones to
    a coordinate. The zones and corresponding y-coordinate are depicted below.


        ohl      | ohc |      ohr
     --------------------------------
          ||=================||
     -------------------------------
          || ghl | ghc | ghr ||
     --------------------------------
      ocl || gcl | gc  | gcr || ocr
     --------------------------------
      oll || gll | glc | glr || olr

      40     45    50    55     60    (y-coordinate of zone)
        44.62               55.38     (y-coordiante of post)
    """
    if (
        _has_tag(raw_event, wyscout_tags.GOAL_LOW_CENTER)
        or _has_tag(raw_event, wyscout_tags.GOAL_CENTER)
        or _has_tag(raw_event, wyscout_tags.GOAL_HIGH_CENTER)
    ):
        return Point(100.0, 50.0)
    if (
        _has_tag(raw_event, wyscout_tags.GOAL_LOW_RIGHT)
        or _has_tag(raw_event, wyscout_tags.GOAL_CENTER_RIGHT)
        or _has_tag(raw_event, wyscout_tags.GOAL_HIGH_RIGHT)
    ):
        return Point(100.0, 55.0)
    if (
        _has_tag(raw_event, wyscout_tags.GOAL_LOW_LEFT)
        or _has_tag(raw_event, wyscout_tags.GOAL_CENTER_LEFT)
        or _has_tag(raw_event, wyscout_tags.GOAL_HIGH_LEFT)
    ):
        return Point(100.0, 45.0)
    if _has_tag(raw_event, wyscout_tags.OUT_HIGH_CENTER) or _has_tag(
        raw_event, wyscout_tags.POST_HIGH_CENTER
    ):
        return Point(100.0, 50.0)
    if (
        _has_tag(raw_event, wyscout_tags.OUT_LOW_RIGHT)
        or _has_tag(raw_event, wyscout_tags.OUT_CENTER_RIGHT)
        or _has_tag(raw_event, wyscout_tags.OUT_HIGH_RIGHT)
    ):
        return Point(100.0, 60.0)
    if (
        _has_tag(raw_event, wyscout_tags.OUT_LOW_LEFT)
        or _has_tag(raw_event, wyscout_tags.OUT_CENTER_LEFT)
        or _has_tag(raw_event, wyscout_tags.OUT_HIGH_LEFT)
    ):
        return Point(100.0, 40.0)
    if (
        _has_tag(raw_event, wyscout_tags.POST_LOW_LEFT)
        or _has_tag(raw_event, wyscout_tags.POST_CENTER_LEFT)
        or _has_tag(raw_event, wyscout_tags.POST_HIGH_LEFT)
    ):
        return Point(100.0, 55.38)
    if (
        _has_tag(raw_event, wyscout_tags.POST_LOW_RIGHT)
        or _has_tag(raw_event, wyscout_tags.POST_CENTER_RIGHT)
        or _has_tag(raw_event, wyscout_tags.POST_HIGH_RIGHT)
    ):
        return Point(100.0, 44.62)
    if _has_tag(raw_event, wyscout_tags.BLOCKED):
        return Point(
            x=float(raw_event["positions"][0]["x"]),
            y=float(raw_event["positions"][0]["y"]),
        )
    return None


def _parse_shot(raw_event: Dict, next_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    qualifiers.extend(_bodypart_qualifiers(raw_event))

    result = None
    if _has_tag(raw_event, 101):
        result = ShotResult.GOAL
    elif _has_tag(raw_event, 2101):
        result = ShotResult.BLOCKED
    elif any(_has_tag(raw_event, tag) for tag in wyscout_tags.SHOT_POST):
        result = ShotResult.POST
    elif any(_has_tag(raw_event, tag) for tag in wyscout_tags.SHOT_OFF_TARGET):
        result = ShotResult.OFF_TARGET
    elif any(_has_tag(raw_event, tag) for tag in wyscout_tags.SHOT_ON_GOAL):
        result = ShotResult.SAVED

    if next_event["eventId"] == wyscout_events.SAVE.EVENT:
        if next_event["subEventId"] == wyscout_events.SAVE.REFLEXES:
            qualifiers.append(GoalkeeperQualifier(GoalkeeperActionType.REFLEX))
        if next_event["subEventId"] == wyscout_events.SAVE.SAVE_ATTEMPT:
            qualifiers.append(
                GoalkeeperQualifier(GoalkeeperActionType.SAVE_ATTEMPT)
            )

    return {
        "result": result,
        "result_coordinates": _create_shot_result_coordinates(raw_event),
        "qualifiers": qualifiers,
    }


def _pass_qualifiers(raw_event) -> List[Qualifier]:
    qualifiers = _generic_qualifiers(raw_event)

    if raw_event["subEventId"] == wyscout_events.PASS.CROSS:
        qualifiers.append(PassQualifier(PassType.CROSS))
    elif raw_event["subEventId"] == wyscout_events.PASS.HAND:
        qualifiers.append(PassQualifier(PassType.HAND_PASS))
        qualifiers.append(BodyPartQualifier(BodyPart.KEEPER_ARM))
    elif raw_event["subEventId"] == wyscout_events.PASS.HEAD:
        qualifiers.append(PassQualifier(PassType.HEAD_PASS))
        qualifiers.append(BodyPartQualifier(BodyPart.HEAD))
    elif raw_event["subEventId"] == wyscout_events.PASS.HIGH:
        qualifiers.append(PassQualifier(PassType.HIGH_PASS))
    elif raw_event["subEventId"] == wyscout_events.PASS.LAUNCH:
        qualifiers.append(PassQualifier(PassType.LAUNCH))
    elif raw_event["subEventId"] == wyscout_events.PASS.SIMPLE:
        qualifiers.append(PassQualifier(PassType.SIMPLE_PASS))
    elif raw_event["subEventId"] == wyscout_events.PASS.SMART:
        qualifiers.append(PassQualifier(PassType.SMART_PASS))

    # If the subevent type did not define the bodypart, we infer it from the tags
    if not any(isinstance(q, BodyPartQualifier) for q in qualifiers):
        qualifiers.extend(_bodypart_qualifiers(raw_event))

    if _has_tag(raw_event, wyscout_tags.HIGH):
        qualifiers.append(PassQualifier(PassType.HIGH_PASS))
    if _has_tag(raw_event, wyscout_tags.THROUGH):
        qualifiers.append(PassQualifier(PassType.THROUGH_BALL))
    if _has_tag(raw_event, wyscout_tags.ASSIST):
        qualifiers.append(PassQualifier(PassType.ASSIST))

    return qualifiers


def _parse_pass(raw_event: Dict, next_event: Dict) -> Dict:
    pass_result = None
    if _has_tag(raw_event, wyscout_tags.ACCURATE):
        pass_result = PassResult.COMPLETE
    elif _has_tag(raw_event, wyscout_tags.NOT_ACCURATE):
        pass_result = PassResult.INCOMPLETE

    receiver_coordinates = None
    if _has_tag(raw_event, wyscout_tags.BLOCKED):
        # blocked passes do not have end coordinates; the start coordinates
        # are used instead
        receiver_coordinates = Point(
            x=float(raw_event["positions"][0]["x"]),
            y=float(raw_event["positions"][0]["y"]),
        )
    elif len(raw_event["positions"]) > 1:
        receiver_coordinates = Point(
            x=float(raw_event["positions"][1]["x"]),
            y=float(raw_event["positions"][1]["y"]),
        )

    if next_event:
        if next_event["eventId"] == wyscout_events.OFFSIDE.EVENT:
            pass_result = PassResult.OFFSIDE
        if next_event["eventId"] == wyscout_events.INTERRUPTION.EVENT:
            if (
                next_event["subEventId"]
                == wyscout_events.INTERRUPTION.BALL_OUT
            ):
                pass_result = PassResult.OUT

    return {
        "result": pass_result,
        "qualifiers": _pass_qualifiers(raw_event),
        "receive_timestamp": None,
        "receiver_player": None,
        "receiver_coordinates": receiver_coordinates,
    }


def _parse_clearance(raw_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    return {"result": None, "qualifiers": qualifiers}


def _parse_goalkeeper_save(raw_event) -> List[Qualifier]:
    qualifiers = _generic_qualifiers(raw_event)
    goalkeeper_qualifiers = []
    if not _has_tag(raw_event, wyscout_tags.GOAL):
        goalkeeper_qualifiers.append(
            GoalkeeperQualifier(value=GoalkeeperActionType.SAVE)
        )
    else:
        goalkeeper_qualifiers.append(
            GoalkeeperQualifier(value=GoalkeeperActionType.SAVE_ATTEMPT)
        )
    if raw_event["subEventId"] == wyscout_events.SAVE.REFLEXES:
        goalkeeper_qualifiers.append(
            GoalkeeperQualifier(value=GoalkeeperActionType.REFLEX)
        )
    qualifiers.extend(goalkeeper_qualifiers)
    return {
        "result": None,
        "qualifiers": qualifiers,
        # start coordinates are stored as inverted end coordinates
        "coordinates": Point(
            x=100.0 - float(raw_event["positions"][1]["x"]),
            y=100.0 - float(raw_event["positions"][1]["y"]),
        ),
    }


def _parse_foul(raw_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)

    if _has_tag(raw_event, wyscout_tags.RED_CARD):
        qualifiers.append(CardQualifier(value=CardType.RED))
    elif _has_tag(raw_event, wyscout_tags.YELLOW_CARD):
        qualifiers.append(CardQualifier(value=CardType.FIRST_YELLOW))
    elif _has_tag(raw_event, wyscout_tags.SECOND_YELLOW_CARD):
        qualifiers.append(CardQualifier(value=CardType.SECOND_YELLOW))

    return {
        "result": None,
        "qualifiers": qualifiers,
    }


def _parse_card(raw_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    card_type = None
    if _has_tag(raw_event, wyscout_tags.RED_CARD):
        card_type = CardType.RED
    elif _has_tag(raw_event, wyscout_tags.YELLOW_CARD):
        card_type = CardType.FIRST_YELLOW
    elif _has_tag(raw_event, wyscout_tags.SECOND_YELLOW_CARD):
        card_type = CardType.SECOND_YELLOW

    return {"result": None, "qualifiers": qualifiers, "card_type": card_type}


def _parse_recovery(raw_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    return {
        "result": None,
        "qualifiers": qualifiers,
    }


def _parse_ball_out(raw_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    return {"result": None, "qualifiers": qualifiers}


def _parse_set_piece(raw_event: Dict, next_event: Dict) -> Dict:
    result = {}
    if raw_event["subEventId"] in wyscout_events.FREE_KICK.PASS_TYPES:
        result = _parse_pass(raw_event, next_event)
        if raw_event["subEventId"] == wyscout_events.FREE_KICK.GOAL_KICK:
            result["qualifiers"].append(
                SetPieceQualifier(SetPieceType.GOAL_KICK)
            )
        elif raw_event["subEventId"] == wyscout_events.FREE_KICK.THROW_IN:
            result["qualifiers"].append(
                SetPieceQualifier(SetPieceType.THROW_IN)
            )
            result["qualifiers"].append(PassQualifier(PassType.HAND_PASS))
        elif raw_event["subEventId"] in [
            wyscout_events.FREE_KICK.FREE_KICK,
            wyscout_events.FREE_KICK.FREE_KICK_CROSS,
        ]:
            result["qualifiers"].append(
                SetPieceQualifier(SetPieceType.FREE_KICK)
            )
        elif raw_event["subEventId"] == wyscout_events.FREE_KICK.CORNER:
            result["qualifiers"].append(
                SetPieceQualifier(SetPieceType.CORNER_KICK)
            )
    elif raw_event["subEventId"] in wyscout_events.FREE_KICK.SHOT_TYPES:
        result = _parse_shot(raw_event, next_event)
        if raw_event["subEventId"] == wyscout_events.FREE_KICK.FREE_KICK_SHOT:
            result["qualifiers"].append(
                SetPieceQualifier(SetPieceType.FREE_KICK)
            )
        elif raw_event["subEventId"] == wyscout_events.FREE_KICK.PENALTY:
            result["qualifiers"].append(
                SetPieceQualifier(SetPieceType.PENALTY)
            )
    else:
        result["qualifiers"] = _generic_qualifiers(raw_event)
    return result


def _parse_interception(raw_event: Dict, next_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    result = InterceptionResult.SUCCESS
    ball_owning_events = (
        wyscout_events.PASS.EVENT,
        wyscout_events.SHOT.EVENT,
    )

    if next_event is not None:
        if next_event["eventId"] == wyscout_events.INTERRUPTION.EVENT:
            if (
                next_event["subEventId"]
                == wyscout_events.INTERRUPTION.BALL_OUT
            ):
                result = InterceptionResult.OUT
        elif raw_event["eventId"] == wyscout_events.PASS.EVENT:
            result = (
                InterceptionResult.LOST
                if _has_tag(raw_event, wyscout_tags.NOT_ACCURATE) is True
                else InterceptionResult.SUCCESS
            )
        # check whether team keeps ball possession
        elif next_event["eventId"] in ball_owning_events:
            if raw_event["teamId"] != next_event["teamId"]:
                result = InterceptionResult.LOST

    return {
        "result": result,
        "qualifiers": qualifiers,
    }


def _parse_duel(raw_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    duel_qualifiers = []

    sub_event_id = raw_event["subEventId"]

    if sub_event_id == wyscout_events.DUEL.AERIAL:
        duel_qualifiers.extend(
            [
                DuelQualifier(value=DuelType.LOOSE_BALL),
                DuelQualifier(value=DuelType.AERIAL),
            ]
        )
    elif sub_event_id in [
        wyscout_events.DUEL.GROUND_ATTACKING,
        wyscout_events.DUEL.GROUND_DEFENDING,
    ]:
        duel_qualifiers.extend([DuelQualifier(value=DuelType.GROUND)])
    elif sub_event_id == wyscout_events.DUEL.GROUND_LOOSE_BALL:
        duel_qualifiers.extend(
            [
                DuelQualifier(value=DuelType.LOOSE_BALL),
                DuelQualifier(value=DuelType.GROUND),
            ]
        )

    if _has_tag(raw_event, wyscout_tags.SLIDING_TACKLE):
        duel_qualifiers.extend([DuelQualifier(value=DuelType.SLIDING_TACKLE)])

    qualifiers.extend(duel_qualifiers)

    result = None
    if _has_tag(raw_event, wyscout_tags.WON):
        result = DuelResult.WON
    elif _has_tag(raw_event, wyscout_tags.LOST):
        result = DuelResult.LOST
    elif _has_tag(raw_event, wyscout_tags.NEUTRAL):
        result = DuelResult.NEUTRAL

    return {"result": result, "qualifiers": qualifiers}


def _players_to_dict(players: List[Player]):
    return {player.player_id: player for player in players}


class WyscoutInputs(NamedTuple):
    event_data: IO[bytes]


class WyscoutDeserializerV2(EventDataDeserializer[WyscoutInputs]):
    @property
    def provider(self) -> Provider:
        return Provider.WYSCOUT

    def deserialize(self, inputs: WyscoutInputs) -> EventDataset:
        transformer = self.get_transformer()

        with performance_logging("load data", logger=logger):
            raw_events = json.load(inputs.event_data)
            for event in raw_events["events"]:
                if "eventId" not in event:
                    event["eventId"] = event["eventName"]
                if "subEventId" not in event:
                    event["subEventId"] = event.get("subEventName")

        periods = []

        with performance_logging("parse data", logger=logger):
            home_team_id, away_team_id = raw_events["teams"].keys()
            home_team = _parse_team(raw_events, home_team_id, Ground.HOME)
            away_team = _parse_team(raw_events, away_team_id, Ground.AWAY)
            teams = {home_team_id: home_team, away_team_id: away_team}
            players = {
                wyId: _players_to_dict(team.players)
                for wyId, team in teams.items()
            }
            game_id = raw_events["events"][0].get("matchId", None)
            if game_id:
                game_id = str(game_id)

            events = []

            for idx, raw_event in enumerate(raw_events["events"]):
                next_event = None
                next_period_id = None
                if (idx + 1) < len(raw_events["events"]):
                    next_event = raw_events["events"][idx + 1]
                    next_period_id = int(
                        next_event["matchPeriod"].replace("H", "")
                    )

                team_id = str(raw_event["teamId"])
                player_id = str(raw_event["playerId"])
                period_id = int(raw_event["matchPeriod"].replace("H", ""))

                if len(periods) == 0 or periods[-1].id != period_id:
                    periods.append(
                        Period(
                            id=period_id,
                            start_timestamp=(
                                timedelta(seconds=0)
                                if len(periods) == 0
                                else periods[-1].end_timestamp
                            ),
                            end_timestamp=None,
                        )
                    )
                if next_period_id != period_id:
                    periods[-1] = replace(
                        periods[-1],
                        end_timestamp=periods[-1].start_timestamp
                        + timedelta(seconds=raw_event["eventSec"]),
                    )

                generic_event_args = {
                    "event_id": str(raw_event["id"]),
                    "raw_event": raw_event,
                    "coordinates": Point(
                        x=float(raw_event["positions"][0]["x"]),
                        y=float(raw_event["positions"][0]["y"]),
                    ),
                    "team": teams[team_id],
                    "player": (
                        players[team_id][player_id]
                        if player_id != INVALID_PLAYER
                        else None
                    ),
                    "ball_owning_team": None,
                    "ball_state": None,
                    "period": periods[-1],
                    "timestamp": timedelta(seconds=raw_event["eventSec"]),
                }

                new_events = []
                if raw_event["eventId"] == wyscout_events.SHOT.EVENT:
                    shot_event_args = _parse_shot(raw_event, next_event)
                    shot_event = self.event_factory.build_shot(
                        **shot_event_args, **generic_event_args
                    )
                    new_events.append(shot_event)
                elif raw_event["eventId"] == wyscout_events.PASS.EVENT:
                    pass_event_args = _parse_pass(raw_event, next_event)
                    pass_event = self.event_factory.build_pass(
                        **pass_event_args, **generic_event_args
                    )
                    new_events.append(pass_event)
                elif raw_event["eventId"] == wyscout_events.FOUL.EVENT:
                    foul_event_args = _parse_foul(raw_event)
                    foul_event = self.event_factory.build_foul_committed(
                        **foul_event_args, **generic_event_args
                    )
                    new_events.append(foul_event)
                    if any(
                        _has_tag(raw_event, tag) for tag in wyscout_tags.CARD
                    ):
                        card_event_args = _parse_card(raw_event)
                        card_event_id = (
                            f"card-{generic_event_args['event_id']}"
                        )
                        card_event = self.event_factory.build_card(
                            **card_event_args,
                            **{
                                **generic_event_args,
                                "event_id": card_event_id,
                            },
                        )
                        new_events.append(card_event)
                elif raw_event["eventId"] == wyscout_events.INTERRUPTION.EVENT:
                    ball_out_event_args = _parse_ball_out(raw_event)
                    ball_out_event = self.event_factory.build_ball_out(
                        **ball_out_event_args, **generic_event_args
                    )
                    new_events.append(ball_out_event)
                elif raw_event["eventId"] == wyscout_events.SAVE.EVENT:
                    goalkeeper_save_args = _parse_goalkeeper_save(raw_event)
                    goalkeeper_save_event = (
                        self.event_factory.build_goalkeeper_event(
                            **{**goalkeeper_save_args, **generic_event_args}
                        )
                    )
                    new_events.append(goalkeeper_save_event)
                elif raw_event["eventId"] == wyscout_events.FREE_KICK.EVENT:
                    set_piece_event_args = _parse_set_piece(
                        raw_event, next_event
                    )
                    if (
                        raw_event["subEventId"]
                        in wyscout_events.FREE_KICK.PASS_TYPES
                    ):
                        fk_pass_event = self.event_factory.build_pass(
                            **set_piece_event_args, **generic_event_args
                        )
                        new_events.append(fk_pass_event)
                    elif (
                        raw_event["subEventId"]
                        in wyscout_events.FREE_KICK.SHOT_TYPES
                    ):
                        fk_shot_event = self.event_factory.build_shot(
                            **set_piece_event_args, **generic_event_args
                        )
                        new_events.append(fk_shot_event)

                elif (
                    raw_event["eventId"] == wyscout_events.OTHERS_ON_BALL.EVENT
                ):
                    if (
                        raw_event["subEventId"]
                        == wyscout_events.OTHERS_ON_BALL.CLEARANCE
                    ):
                        clearance_event_args = _parse_clearance(raw_event)
                        clearance_event = self.event_factory.build_clearance(
                            **clearance_event_args,
                            **generic_event_args,
                        )
                        new_events.append(clearance_event)
                    elif (
                        raw_event["subEventId"]
                        == wyscout_events.OTHERS_ON_BALL.TOUCH
                    ) & (_has_tag(raw_event, wyscout_tags.MISSED_BALL)):
                        miscontrol_event_args = {
                            "result": None,
                            "qualifiers": _generic_qualifiers(raw_event),
                        }
                        miscontrol_event = self.event_factory.build_miscontrol(
                            **miscontrol_event_args,
                            **generic_event_args,
                        )
                        new_events.append(miscontrol_event)
                    else:
                        recovery_event_args = _parse_recovery(raw_event)
                        recovery_event = self.event_factory.build_recovery(
                            **recovery_event_args, **generic_event_args
                        )
                        new_events.append(recovery_event)
                elif raw_event["eventId"] == wyscout_events.DUEL.EVENT:
                    duel_event_args = _parse_duel(raw_event)
                    duel_event = self.event_factory.build_duel(
                        **duel_event_args, **generic_event_args
                    )
                    new_events.append(duel_event)
                elif raw_event["eventId"] not in [
                    wyscout_events.SAVE.EVENT,
                    wyscout_events.OFFSIDE.EVENT,
                ]:
                    # The events SAVE and OFFSIDE are already merged with PASS and SHOT events
                    qualifiers = _generic_qualifiers(raw_event)
                    generic_event = self.event_factory.build_generic(
                        result=None,
                        qualifiers=qualifiers,
                        **generic_event_args,
                    )
                    new_events.append(generic_event)

                # Wyscout v2 does not have a separate event type for
                # interceptions. Interceptions are recorded by adding a tag to
                # the next pass, touch or duel. Therefore, we convert events
                # with this tag to an interception.
                if _has_tag(raw_event, wyscout_tags.INTERCEPTION):
                    interception_event_args = _parse_interception(
                        raw_event, next_event
                    )

                    for i, new_event in enumerate(list(new_events)):
                        if new_event.event_type == EventType.DUEL:
                            # when DuelEvent is interception, we need to
                            # overwrite this and the previous DuelEvent
                            events = events[:-1]
                            new_events[
                                i
                            ] = self.event_factory.build_interception(
                                **interception_event_args,
                                **generic_event_args,
                            )
                        elif new_event.event_type in [
                            EventType.RECOVERY,
                            EventType.MISCONTROL,
                        ]:
                            # replace touch events
                            new_events[
                                i
                            ] = self.event_factory.build_interception(
                                **interception_event_args,
                                **generic_event_args,
                            )
                        elif new_event.event_type in [
                            EventType.PASS,
                            EventType.CLEARANCE,
                        ]:
                            # insert an interception event before interception passes
                            generic_event_args[
                                "event_id"
                            ] = f"interception-{generic_event_args['event_id']}"
                            interception_event = (
                                self.event_factory.build_interception(
                                    **interception_event_args,
                                    **generic_event_args,
                                )
                            )
                            new_events.insert(i, interception_event)

                for new_event in new_events:
                    if self.should_include_event(new_event):
                        events.append(transformer.transform_event(new_event))

        metadata = Metadata(
            teams=[home_team, away_team],
            periods=periods,
            pitch_dimensions=transformer.get_to_coordinate_system().pitch_dimensions,
            score=None,
            frame_rate=None,
            orientation=Orientation.ACTION_EXECUTING_TEAM,
            flags=None,
            provider=Provider.WYSCOUT,
            coordinate_system=transformer.get_to_coordinate_system(),
            game_id=game_id,
        )

        return EventDataset(metadata=metadata, records=events)



"""Base class for all Opta(-derived) event stream parsers.

A parser reads a single data file and should extend the 'OptaParser' class to
extract data about players, teams and events that is encoded in the file.
"""

import json
from typing import Tuple, List, Optional, IO, Dict

from lxml import objectify

from kloppy.domain import Team, Score, Period, PositionType

from datetime import datetime
from dataclasses import dataclass, field


position_types_mapping: Dict[str, PositionType] = {
    "Goalkeeper": PositionType.Goalkeeper,
    "Defender": PositionType.Defender,
    "Midfielder": PositionType.Midfielder,
    "Striker": PositionType.Attacker,
    "Substitute": PositionType.Unknown,
}


@dataclass
class OptaEvent:
    """A raw Opta event."""

    id: str
    event_id: int
    type_id: int
    period_id: int
    time_min: int
    time_sec: int
    x: float
    y: float
    timestamp: datetime
    last_modified: datetime
    contestant_id: Optional[str] = None
    player_id: Optional[str] = None
    outcome: Optional[int] = None
    qualifiers: Dict[int, str] = field(default_factory=dict)


class OptaParser:
    """Extract data from an Opta data stream.

    Args:
        feed : The data stream of a game to parse.
    """

    def __init__(self, feed: IO[bytes]) -> None:
        raise NotImplementedError

    def extract_periods(self) -> List[Period]:
        """Return the periods of the game."""
        raise NotImplementedError

    def extract_score(self) -> Optional[Score]:
        """Return the score of the game."""
        return None

    def extract_date(self) -> Optional[datetime]:
        """Return the date of the game."""
        return None

    def extract_game_week(self) -> Optional[str]:
        """Return the game_week of the game."""
        return None

    def extract_game_id(self) -> Optional[str]:
        """Return the game_id of the game."""
        return None

    def extract_lineups(self) -> Tuple[Team, Team]:
        """Return the home and away team."""
        raise NotImplementedError

    def extract_events(self) -> List[OptaEvent]:
        """Return all events."""
        raise NotImplementedError


class OptaJSONParser(OptaParser):
    """Extract data from an Opta JSON data stream.

    Args:
        feed : The data stream of a game to parse.

    Attributes:
        root : The root of the JSON data.
    """

    def __init__(self, feed: IO[bytes]) -> None:
        self.root = json.load(feed)


class OptaXMLParser(OptaParser):
    """Extract data from an Opta XML data stream.

    Args:
        feed : The data stream of a game to parse.

    Attributes:
        root : The root of the XML tree.
    """

    def __init__(self, feed: IO[bytes]) -> None:
        self.root = objectify.fromstring(feed.read())



import json
from datetime import timedelta
from typing import IO, Iterator, List, Tuple

from kloppy.domain import (
    BallState,
    Frame,
    Period,
    PlayerData,
    Point,
    Point3D,
    Team,
)
from kloppy.domain.services.frame_factory import create_frame
from kloppy.exceptions import DeserializationError

from .base import TracabDataParser


class TracabJSONParser(TracabDataParser):
    def __init__(
        self,
        feed: IO[bytes],
        periods: List[Period],
        teams: Tuple[Team, Team],
        frame_rate: int,
    ) -> None:
        self.root = json.load(feed)
        self.periods = periods
        self.teams = teams
        self.frame_rate = frame_rate

    def extract_frames(
        self, sample_rate: float, only_alive: bool
    ) -> Iterator[Frame]:
        raw_data = self.root["FrameData"]

        n = 0
        sample = 1.0 / sample_rate

        for frame in raw_data:
            if only_alive and frame["BallPosition"][0]["BallStatus"] == "Dead":
                continue

            frame_id = frame["FrameCount"]
            for period in self.periods:
                assert isinstance(
                    period.start_timestamp, timedelta
                ), "The period's start_timestamp should be a relative time (i.e., a timedelta object)"
                assert isinstance(
                    period.end_timestamp, timedelta
                ), "The period's start_timestamp should be a relative time (i.e., a timedelta object)"

                if (
                    period.start_timestamp
                    <= timedelta(seconds=frame_id / self.frame_rate)
                    <= period.end_timestamp
                ):
                    if n % sample == 0:
                        frame = self._parse_frame(period, frame)
                        yield frame
                    n += 1

    def _parse_frame(self, period, raw_frame):
        frame_id = raw_frame["FrameCount"]
        raw_players_data = raw_frame["PlayerPositions"]
        raw_ball_position = raw_frame["BallPosition"][0]

        players_data = {}
        for player_data in raw_players_data:
            if player_data["Team"] == 1:
                team = self.teams[0]
            elif player_data["Team"] == 0:
                team = self.teams[1]
            elif player_data["Team"] in (-1, 3, 4):
                continue
            else:
                raise DeserializationError(
                    f"Unknown Player Team ID: {player_data['Team']}"
                )

            jersey_no = player_data["JerseyNumber"]
            x = player_data["X"]
            y = player_data["Y"]
            speed = player_data["Speed"]

            player = team.get_player_by_jersey_number(jersey_no)
            if player:
                players_data[player] = PlayerData(
                    coordinates=Point(x, y), speed=speed
                )
            else:
                # continue
                raise DeserializationError(
                    f"Player not found for player jersey no {jersey_no} of team: {team.name}"
                )

        ball_x = raw_ball_position["X"]
        ball_y = raw_ball_position["Y"]
        ball_z = raw_ball_position["Z"]
        ball_speed = raw_ball_position["Speed"]
        if raw_ball_position["BallOwningTeam"] == "H":
            ball_owning_team = self.teams[0]
        elif raw_ball_position["BallOwningTeam"] == "A":
            ball_owning_team = self.teams[1]
        else:
            raise DeserializationError(
                f"Unknown ball owning team: {raw_ball_position['BallOwningTeam']}"
            )
        if raw_ball_position["BallStatus"] == "Alive":
            ball_state = BallState.ALIVE
        elif raw_ball_position["BallStatus"] == "Dead":
            ball_state = BallState.DEAD
        else:
            raise DeserializationError(
                f"Unknown ball state: {raw_ball_position['BallStatus']}"
            )

        frame = create_frame(
            frame_id=frame_id,
            timestamp=timedelta(seconds=frame_id / self.frame_rate)
            - period.start_timestamp,
            ball_coordinates=Point3D(ball_x, ball_y, ball_z),
            ball_state=ball_state,
            ball_owning_team=ball_owning_team,
            ball_speed=ball_speed,
            players_data=players_data,
            period=period,
            other_data={},
        )

        return frame



import json
from typing import Optional, Union

from kloppy.domain import TrackingDataset
from kloppy.exceptions import DeserializationError
from kloppy.infra.serializers.tracking.skillcorner import (
    SkillCornerDeserializer,
    SkillCornerInputs,
)
from kloppy.io import FileLike, open_as_file


def load(
    meta_data: FileLike,
    raw_data: FileLike,
    sample_rate: Optional[float] = None,
    limit: Optional[int] = None,
    coordinates: Optional[str] = None,
    include_empty_frames: Optional[bool] = False,
    data_version: Optional[str] = None,
) -> TrackingDataset:
    """
    Load SkillCorner broadcast tracking data.

    Args:
        meta_data: A json feed containing the meta data.
        raw_data: A json feed containing the raw tracking data.
        sample_rate: Sample the data at a specific rate.
        limit: Limit the number of frames to load to the first `limit` frames.
        coordinates: The coordinate system to use.
        include_empty_frames: Include frames in which no objects were tracked.
        data_version: Specify the input data version.

    Returns:
        The parsed tracking data.
    """
    if data_version not in ["V2", "V3", None]:
        raise ValueError(
            f"data_version must be either 'V2', 'V3'. Provided: {data_version}"
        )
    if not data_version:
        data_version = identify_data_version(raw_data)
    deserializer = SkillCornerDeserializer(
        sample_rate=sample_rate,
        limit=limit,
        coordinate_system=coordinates,
        include_empty_frames=include_empty_frames,
        data_version=data_version,
    )
    with open_as_file(meta_data) as meta_data_fp, open_as_file(
        raw_data
    ) as raw_data_fp:
        return deserializer.deserialize(
            inputs=SkillCornerInputs(
                meta_data=meta_data_fp, raw_data=raw_data_fp
            )
        )


def load_open_data(
    match_id: Union[str, int] = "4039",
    sample_rate: Optional[float] = None,
    limit: Optional[int] = None,
    coordinates: Optional[str] = None,
    include_empty_frames: Optional[bool] = False,
) -> TrackingDataset:
    """
    Load SkillCorner open data.

    This function loads tracking data directly from the SkillCorner open data
    GitHub repository.

    Args:
        match_id: The id of the match to load data for.
        sample_rate: Sample the data at a specific rate.
        limit: Limit the number of frames to load to the first `limit` frames.
        coordinates: The coordinate system to use.
        include_empty_frames: Include frames in which no objects were tracked.

    Returns:
        The parsed tracking data.
    """
    return load(
        meta_data=f"https://raw.githubusercontent.com/SkillCorner/opendata/master/data/matches/{match_id}/match_data.json",
        raw_data=f"https://raw.githubusercontent.com/SkillCorner/opendata/master/data/matches/{match_id}/structured_data.json",
        sample_rate=sample_rate,
        limit=limit,
        coordinates=coordinates,
        include_empty_frames=include_empty_frames,
    )


def identify_data_version(raw_data: FileLike) -> str:
    """
    Identify the data version of the SkillCorner data based on the raw data file.
    Supports both JSONL and JSON formats.
    """
    with open_as_file(raw_data) as raw_data_fp:
        # Extract the first few bytes
        start_byte = raw_data_fp.read(1)
        raw_data_fp.seek(0)

        # Check if it starts with '{' or '['
        if start_byte == b"[":
            # It's a JSON array
            first_line = json.load(raw_data_fp)[0]
        elif start_byte == b"{":
            # It's a JSONL file
            first_line = json.loads(raw_data_fp.readline().strip())
        else:
            raise DeserializationError("Could not determine raw data format")

        # JSONL case: first line is a JSON object
        if "data" in first_line:
            return "V2"
        elif "player_data" in first_line:
            return "V3"

        raise ValueError("Unexpected SkillCorner raw data format")



import json
import logging
import warnings
from dataclasses import replace
from datetime import datetime, timedelta, timezone
from enum import Enum
from typing import Dict, List, Optional

from kloppy.domain import (
    BodyPart,
    BodyPartQualifier,
    CardType,
    CarryResult,
    CounterAttackQualifier,
    DuelQualifier,
    DuelResult,
    DuelType,
    EventDataset,
    ExpectedGoals,
    FormationType,
    GoalkeeperActionType,
    GoalkeeperQualifier,
    Ground,
    InterceptionResult,
    Metadata,
    Orientation,
    PassQualifier,
    PassResult,
    PassType,
    Period,
    Player,
    Point,
    Point3D,
    PositionType,
    PostShotExpectedGoals,
    Provider,
    Qualifier,
    SetPieceQualifier,
    SetPieceType,
    ShotResult,
    TakeOnResult,
    Team,
)
from kloppy.exceptions import DeserializationError, DeserializationWarning
from kloppy.utils import performance_logging

from ..deserializer import EventDataDeserializer
from .deserializer_v2 import WyscoutInputs

logger = logging.getLogger(__name__)


INVALID_PLAYER = "0"

formations = {
    "4-4-2": FormationType.FOUR_FOUR_TWO,
    "4-4-1-1": FormationType.FOUR_FOUR_ONE_ONE,
    "4-3-2-1": FormationType.FOUR_THREE_TWO_ONE,
    "4-2-3-1": FormationType.FOUR_TWO_THREE_ONE,
    "4-1-4-1": FormationType.FOUR_ONE_FOUR_ONE,
    "4-1-3-2": FormationType.FOUR_ONE_THREE_TWO,
    "4-3-1-2": FormationType.FOUR_THREE_ONE_TWO,
    "4-3-3": FormationType.FOUR_THREE_THREE,
    "4-5-1": FormationType.FOUR_FIVE_ONE,
    "4-2-2-2": FormationType.FOUR_TWO_TWO_TWO,
    "4-2-1-3": FormationType.FOUR_TWO_ONE_THREE,
    "3-4-3": FormationType.THREE_FOUR_THREE,
    "3-4-1-2": FormationType.THREE_FOUR_ONE_TWO,
    "3-4-2-1": FormationType.THREE_FOUR_TWO_ONE,
    "3-5-2": FormationType.THREE_FIVE_TWO,
    "3-5-1-1": FormationType.THREE_FIVE_ONE_ONE,
    "5-3-2": FormationType.FIVE_THREE_TWO,
    "5-4-1": FormationType.FIVE_FOUR_ONE,
    "3-3-3-1": FormationType.THREE_THREE_THREE_ONE,
    "3-2-3-2": FormationType.THREE_TWO_THREE_TWO,
}

position_types_mapping: Dict[str, PositionType] = {
    "GK": PositionType.Goalkeeper,
    "LB": PositionType.LeftBack,
    "LWB": PositionType.LeftWingBack,
    "LB5": PositionType.LeftBack,
    "LCB": PositionType.LeftCenterBack,
    "LCB3": PositionType.LeftCenterBack,
    "CB": PositionType.CenterBack,
    "RCB": PositionType.RightCenterBack,
    "RCB3": PositionType.RightCenterBack,
    "RB": PositionType.RightBack,
    "RWB": PositionType.RightWingBack,
    "RB5": PositionType.RightBack,
    "LW": PositionType.LeftWing,
    "LAMF": PositionType.LeftAttackingMidfield,
    "LCMF3": PositionType.LeftCentralMidfield,
    "LCMF": PositionType.LeftCentralMidfield,
    "DMF": PositionType.DefensiveMidfield,
    "LDMF": PositionType.LeftDefensiveMidfield,
    "RDMF": PositionType.RightDefensiveMidfield,
    "RCMF3": PositionType.RightCentralMidfield,
    "RCMF": PositionType.RightCentralMidfield,
    "RAMF": PositionType.RightAttackingMidfield,
    "RW": PositionType.RightWing,
    "AMF": PositionType.AttackingMidfield,
    "LWF": PositionType.LeftForward,
    "CF": PositionType.Striker,
    "SS": PositionType.Striker,
    "RWF": PositionType.RightForward,
}


def _flip_point(point: Point) -> Point:
    return Point(x=100 - point.x, y=100 - point.y)


class ShotZoneResults(Enum):
    """
     Wyscout does not provide end-coordinates of shots. Instead shots on goal
     are tagged with a zone. The zones and corresponding y and z-coordinates are depicted below.

                                        (z-coordinate of zone or post)

         olt      | ot |      otr          3.5
      --------------------------------
           ||=================||           2.77
      -------------------------------
           || gtl | gt | gtr ||            2
      --------------------------------
       ol || gl | gc  | gr || or           1
      --------------------------------
       olb || glb  | gb | gbr || orb       0

       40     45    50    55     60    (y-coordinate of zone)
         44.62               55.38     (y-coordinate of post)

    Attributes:
       code (str): The string identifier of the shot result.
       point (Point3D): The predefined Point3D associated with the shot result.
    """

    GOAL_BOTTOM_LEFT = ("glb", Point3D(100, 45, 0))
    GOAL_BOTTOM_CENTER = ("gb", Point3D(100, 50, 0))
    GOAL_BOTTOM_RIGHT = ("gbr", Point3D(100, 55, 0))
    GOAL_CENTER_LEFT = ("gl", Point3D(100, 45, 1))
    GOAL_CENTER = ("gc", Point3D(100, 50, 1))
    GOAL_CENTER_RIGHT = ("gr", Point3D(100, 55, 1))
    GOAL_TOP_LEFT = ("gtl", Point3D(100, 45, 2))
    GOAL_TOP_CENTER = ("gt", Point3D(100, 50, 2))
    GOAL_TOP_RIGHT = ("gtr", Point3D(100, 55, 2))
    POST_BOTTOM_LEFT = ("plb", Point3D(100, 44.62, 0))
    POST_BOTTOM_RIGHT = ("pbr", Point3D(100, 55.38, 0))
    OUT_BOTTOM_LEFT = ("olb", Point3D(100, 40, 0))
    OUT_BOTTOM_RIGHT = ("obr", Point3D(100, 60, 0))
    POST_LEFT = ("pl", Point3D(100, 44.62, 1))
    POST_RIGHT = ("pr", Point3D(100, 55.38, 1))
    OUT_LEFT = ("ol", Point3D(100, 40, 1))
    OUT_RIGHT = ("or", Point3D(100, 60, 1))
    POST_TOP_LEFT = ("ptl", Point3D(100, 44.62, 2))
    POST_TOP_RIGHT = ("ptr", Point3D(100, 55.38, 2))
    OUT_TOP_LEFT = ("otl", Point3D(100, 40, 3.5))
    OUT_TOP_RIGHT = ("otr", Point3D(100, 60, 3.5))
    POST_TOP = ("pt", Point3D(100, 50, 2.77))
    OUT_TOP = ("ot", Point3D(100, 50, 3.5))
    BLOCKED = ("bc", None)

    def __init__(self, code: str, point: Point3D):
        self.code = code
        self.point = point

    @classmethod
    def point_from_code(cls, code: str) -> Optional[Point3D]:
        """Retrieves the enum member from a string code."""
        for zone in cls:
            if zone.code == code:
                return zone.point

        warnings.warn(f"Invalid shot zone code: {code}")


def _parse_team(raw_events, wyId: str, ground: Ground) -> Team:
    # Get the first formation description
    first_period_formation_info = raw_events["formations"][wyId]["1H"]
    first_formation_descr = next(iter(first_period_formation_info.values()))
    formation_str, formation_info = next(iter(first_formation_descr.items()))

    # Extract the formation and players' positions
    starting_formation = formations.get(formation_str)
    starting_players_positions = {
        player_id: position_types_mapping.get(
            player_info["position"].upper(), PositionType.Unknown
        )
        for player_descr in formation_info["players"]
        for player_id, player_info in player_descr.items()
    }

    team = Team(
        team_id=wyId,
        name=raw_events["teams"][wyId]["team"]["officialName"],
        ground=ground,
        starting_formation=starting_formation,
    )

    players_info = {}
    for player in raw_events["players"][wyId]:
        player_id = str(player["player"]["wyId"])
        player_first_name = player["player"]["firstName"]
        player_last_name = player["player"]["lastName"]
        players_info[player_id] = {
            "first_name": player_first_name,
            "last_name": player_last_name,
        }

    team_formation = raw_events["match"]["teamsData"][wyId]["formation"]
    team_players = team_formation["lineup"] + team_formation["bench"]
    for player in team_players:
        player_id = str(player["playerId"])
        starting_position = starting_players_positions.get(player_id)
        player_info = players_info[player_id]
        team.players.append(
            Player(
                player_id=player_id,
                team=team,
                jersey_no=player["shirtNumber"],
                first_name=player_info["first_name"],
                last_name=player_info["last_name"],
                starting=starting_position is not None,
                starting_position=starting_position,
            )
        )

    return team


def _create_shot_result_coordinates(raw_event: Dict) -> Optional[Point]:
    """Estimate the shot end location from the Wyscout tags."""
    if raw_event["shot"]["goalZone"] == ShotZoneResults.BLOCKED.code:
        return Point(
            x=float(raw_event["location"]["x"]),
            y=float(raw_event["location"]["y"]),
        )

    else:
        result_coordinates = ShotZoneResults.point_from_code(
            raw_event["shot"]["goalZone"]
        )
        return result_coordinates


def _generic_qualifiers(raw_event: Dict) -> List[Qualifier]:
    qualifiers: List[Qualifier] = []

    counter_attack_qualifier = CounterAttackQualifier(False)
    if raw_event["possession"]:
        if "counterattack" in raw_event["possession"]["types"]:
            counter_attack_qualifier = CounterAttackQualifier(True)
    qualifiers.append(counter_attack_qualifier)

    return qualifiers


def _parse_shot(raw_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    if raw_event["type"]["primary"] == "own_goal":
        result = ShotResult.OWN_GOAL
    elif raw_event["shot"]["isGoal"] is True:
        result = ShotResult.GOAL
    elif raw_event["shot"]["onTarget"] is True:
        result = ShotResult.SAVED
    elif raw_event["shot"]["goalZone"] == "bc":
        result = ShotResult.BLOCKED
    else:
        result = ShotResult.OFF_TARGET

    if result != ShotResult.OWN_GOAL:
        result_coordinates = _create_shot_result_coordinates(raw_event)
        if raw_event["shot"]["bodyPart"] == "head_or_other":
            qualifiers.append(BodyPartQualifier(value=BodyPart.HEAD))
        elif raw_event["shot"]["bodyPart"] == "left_foot":
            qualifiers.append(BodyPartQualifier(value=BodyPart.LEFT_FOOT))
        elif raw_event["shot"]["bodyPart"] == "right_foot":
            qualifiers.append(BodyPartQualifier(value=BodyPart.RIGHT_FOOT))

        statistics = []
        for statistic_cls, prop_name in {
            ExpectedGoals: "xg",
            PostShotExpectedGoals: "postShotXg",
        }.items():
            value = raw_event["shot"].get(prop_name, None)
            if value is not None:
                statistics.append(statistic_cls(value=value))
    else:
        result_coordinates = None
        statistics = []

    return {
        "result": result,
        "result_coordinates": result_coordinates,
        "qualifiers": qualifiers,
        "statistics": statistics,
    }


def _check_secondary_event_types(
    raw_event, secondary_event_types_values: List[str]
) -> bool:
    return any(
        secondary_event_types in secondary_event_types_values
        for secondary_event_types in raw_event["type"]["secondary"]
    )


def _pass_qualifiers(raw_event) -> List[Qualifier]:
    qualifiers = _generic_qualifiers(raw_event)

    qualifier_mapping = {
        PassType.CROSS: ["cross", "cross_blocked"],
        PassType.HAND_PASS: ["hand_pass"],
        PassType.HEAD_PASS: ["head_pass"],
        PassType.SMART_PASS: ["smart_pass"],
        PassType.SHOT_ASSIST: ["shot_assist"],
        PassType.ASSIST: ["assist"],
        PassType.THROUGH_BALL: ["through_pass"],
    }

    for pass_type, secondary_event_types_values in qualifier_mapping.items():
        if _check_secondary_event_types(
            raw_event, secondary_event_types_values
        ):
            qualifiers.append(PassQualifier(pass_type))

    if raw_event["pass"].get("height") == "high":
        qualifiers.append(PassQualifier(PassType.HIGH_PASS))

    return qualifiers


def _parse_pass(raw_event: Dict, next_event: Dict, team: Team) -> Dict:
    pass_result = None
    receiver_player = None
    if len(raw_event["pass"]["endLocation"]) > 1:
        receiver_coordinates = Point(
            x=float(raw_event["pass"]["endLocation"]["x"]),
            y=float(raw_event["pass"]["endLocation"]["y"]),
        )
    else:
        receiver_coordinates = None

    if raw_event["pass"]["accurate"] is True:
        pass_result = PassResult.COMPLETE
        receiver_player = team.get_player_by_id(
            raw_event["pass"]["recipient"]["id"]
        )
    elif raw_event["pass"]["accurate"] is False:
        pass_result = PassResult.INCOMPLETE

    if raw_event["pass"].get("height") == "blocked":
        receiver_coordinates = None

    if next_event:
        if next_event["type"]["primary"] == "offside":
            pass_result = PassResult.OFFSIDE
        if next_event["type"]["primary"] == "game_interruption":
            if "ball_out" in next_event["type"]["secondary"]:
                pass_result = PassResult.OUT
        # Set end coordinates of blocked pass to start coordinates of next event if it is not a game interruption
        if raw_event["pass"].get("height") == "blocked":
            next_event_location = Point(
                x=float(next_event["location"]["x"]),
                y=float(next_event["location"]["y"]),
            )
            if team.team_id == str(next_event["team"]["id"]):
                receiver_coordinates = next_event_location
            else:
                receiver_coordinates = _flip_point(next_event_location)

    return {
        "result": pass_result,
        "qualifiers": _pass_qualifiers(raw_event),
        "receive_timestamp": None,
        "receiver_player": receiver_player,
        "receiver_coordinates": receiver_coordinates,
    }


def _parse_foul(raw_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    return {
        "result": None,
        "qualifiers": qualifiers,
    }


def _parse_card(raw_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    card_type = None
    if _check_secondary_event_types(raw_event, ["yellow_card"]):
        card_type = CardType.FIRST_YELLOW
    elif _check_secondary_event_types(raw_event, ["red_card"]):
        card_type = CardType.RED

    return {"result": None, "qualifiers": qualifiers, "card_type": card_type}


def _parse_recovery(raw_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    return {
        "result": None,
        "qualifiers": qualifiers,
    }


def _parse_clearance(raw_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    return {
        "result": None,
        "qualifiers": qualifiers,
    }


def _parse_interception(raw_event: Dict, next_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    result = InterceptionResult.SUCCESS

    if next_event is not None:
        is_game_interruption = (
            next_event["type"]["primary"] == "game_interruption"
        )
        is_ball_out = "ball_out" in next_event["type"]["secondary"]
        is_loss = "loss" in raw_event["type"]["secondary"]
        is_pass_loss = (
            "pass" in raw_event["type"]["secondary"]
            and raw_event["pass"]["accurate"] is False
        )
        is_possession_loss = (
            next_event["possession"] is not None
            and raw_event["team"]["id"]
            != next_event["possession"]["team"]["id"]
        )

        if is_game_interruption and is_ball_out:
            result = InterceptionResult.OUT
        elif is_loss or is_pass_loss or is_possession_loss:
            result = InterceptionResult.LOST

    return {
        "result": result,
        "qualifiers": qualifiers,
    }


def _parse_carry(raw_event: Dict, next_event: Dict, start_ts: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    carry_info = raw_event["carry"]
    end_coordinates = Point(
        x=float(carry_info["endLocation"]["x"]),
        y=float(carry_info["endLocation"]["y"]),
    )

    if next_event is not None:
        period_id = _parse_period_id(next_event["matchPeriod"])
        end_timestamp = _create_timestamp_timedelta(
            next_event, start_ts, period_id
        )
    else:
        period_id = _parse_period_id(raw_event["matchPeriod"])
        end_timestamp = _create_timestamp_timedelta(
            raw_event, start_ts, period_id
        )

    return {
        "result": CarryResult.COMPLETE,
        "qualifiers": qualifiers,
        "end_coordinates": end_coordinates,
        "end_timestamp": end_timestamp,
    }


def _parse_goalkeeper_save(raw_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)

    goalkeeper_qualifiers = []
    if "save" in raw_event["type"]["secondary"]:
        goalkeeper_qualifiers.append(
            GoalkeeperQualifier(value=GoalkeeperActionType.SAVE)
        )

    if "save_with_reflex" == "save_with_reflex":
        goalkeeper_qualifiers.append(
            GoalkeeperQualifier(value=GoalkeeperActionType.REFLEX)
        )
    qualifiers.extend(goalkeeper_qualifiers)

    return {"result": None, "qualifiers": qualifiers}


def _parse_ball_out(raw_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    return {"result": None, "qualifiers": qualifiers}


def _parse_set_piece(raw_event: Dict, next_event: Dict, team: Team) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    result = {}

    # Pass set pieces
    if raw_event["type"]["primary"] == "goal_kick":
        qualifiers.append(SetPieceQualifier(SetPieceType.GOAL_KICK))
        result = _parse_pass(raw_event, next_event, team)
    elif raw_event["type"]["primary"] == "throw_in":
        qualifiers.append(SetPieceQualifier(SetPieceType.THROW_IN))
        qualifiers.append(PassQualifier(PassType.HAND_PASS))
        result = _parse_pass(raw_event, next_event, team)
    elif (
        raw_event["type"]["primary"] == "free_kick"
    ) and "free_kick_shot" not in raw_event["type"]["secondary"]:
        qualifiers.append(SetPieceQualifier(SetPieceType.FREE_KICK))
        result = _parse_pass(raw_event, next_event, team)
    elif (
        raw_event["type"]["primary"] == "corner"
    ) and "shot" not in raw_event["type"]["secondary"]:
        qualifiers.append(SetPieceQualifier(SetPieceType.CORNER_KICK))
        result = _parse_pass(raw_event, next_event, team)
    # Shot set pieces
    elif (
        raw_event["type"]["primary"] == "free_kick"
    ) and "free_kick_shot" in raw_event["type"]["secondary"]:
        qualifiers.append(SetPieceQualifier(SetPieceType.FREE_KICK))
        result = _parse_shot(raw_event)
    elif (raw_event["type"]["primary"] == "corner") and "shot" in raw_event[
        "type"
    ]["secondary"]:
        qualifiers.append(SetPieceQualifier(SetPieceType.CORNER_KICK))
        result = _parse_shot(raw_event)
    elif raw_event["type"]["primary"] == "penalty":
        qualifiers.append(SetPieceQualifier(SetPieceType.PENALTY))
        result = _parse_shot(raw_event)

    result["qualifiers"] = qualifiers
    return result


def _parse_take_on(raw_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    result = None
    if "offensive_duel" in raw_event["type"]["secondary"]:
        if raw_event["groundDuel"]["keptPossession"]:
            result = TakeOnResult.COMPLETE
        else:
            result = TakeOnResult.INCOMPLETE
    elif "defensive_duel" in raw_event["type"]["secondary"]:
        if raw_event["groundDuel"]["recoveredPossession"]:
            result = TakeOnResult.COMPLETE
        else:
            result = TakeOnResult.INCOMPLETE
    elif "aerial_duel" in raw_event["type"]["secondary"]:
        if raw_event["aerialDuel"]["firstTouch"]:
            result = TakeOnResult.COMPLETE
        else:
            result = TakeOnResult.INCOMPLETE

    return {"result": result, "qualifiers": qualifiers}


def _parse_duel(raw_event: Dict) -> Dict:
    qualifiers = _generic_qualifiers(raw_event)
    duel_qualifiers = []
    secondary_types = raw_event["type"]["secondary"]

    if "ground_duel" in secondary_types:
        duel_qualifiers.append(DuelQualifier(value=DuelType.GROUND))
    elif "aerial_duel" in secondary_types:
        duel_qualifiers.extend(
            [
                DuelQualifier(value=DuelType.LOOSE_BALL),
                DuelQualifier(value=DuelType.AERIAL),
            ]
        )
    else:
        if (
            "loose_ball_duel" in secondary_types
            and "sliding_tackle" in secondary_types
        ):
            duel_qualifiers.extend(
                [
                    DuelQualifier(value=DuelType.GROUND),
                    DuelQualifier(value=DuelType.LOOSE_BALL),
                    DuelQualifier(value=DuelType.SLIDING_TACKLE),
                ]
            )
        elif "loose_ball_duel" in secondary_types:
            duel_qualifiers.extend(
                [
                    DuelQualifier(value=DuelType.GROUND),
                    DuelQualifier(value=DuelType.LOOSE_BALL),
                ]
            )
        elif "sliding_tackle" in secondary_types:
            duel_qualifiers.extend(
                [
                    DuelQualifier(value=DuelType.GROUND),
                    DuelQualifier(value=DuelType.SLIDING_TACKLE),
                ]
            )

    qualifiers.extend(duel_qualifiers)

    if (
        "offensive_duel" in secondary_types
        and raw_event["groundDuel"]["keptPossession"]
    ):
        result = DuelResult.WON
    elif (
        "defensive_duel" in secondary_types
        and raw_event["groundDuel"]["recoveredPossession"]
    ):
        result = DuelResult.WON
    elif (
        "aerial_duel" in secondary_types
        and raw_event["aerialDuel"]["firstTouch"]
    ):
        result = DuelResult.WON
    else:
        result = DuelResult.LOST

    return {"result": result, "qualifiers": qualifiers}


def _create_timestamp_timedelta(
    raw_event: Dict, start_ts: Dict, period_id: int
) -> timedelta:
    time_delta = (
        timedelta(
            seconds=float(raw_event["second"] + raw_event["minute"] * 60)
        )
        - start_ts[period_id]
    )

    return time_delta


def _validate_formation(formation: str) -> bool:
    formation_parts = formation.split("-")
    formation_ints = [int(part) for part in formation_parts]
    total_players = sum(formation_ints)

    return total_players == 10


def _get_team_formation(team_formation):
    return (
        formations[team_formation]
        if _validate_formation(team_formation)
        else FormationType.UNKNOWN
    )


def get_home_away_team_formation(event, team):
    team_formation = event["team"]["formation"]
    team_id = str(event["team"]["id"])
    opponent_formation = event["opponentTeam"]["formation"]
    opponent_team_id = str(event["opponentTeam"]["id"])

    if team.ground == Ground.HOME:
        home_team_formation = _get_team_formation(team_formation)
        away_team_formation = _get_team_formation(opponent_formation)
    elif team.ground == Ground.AWAY:
        away_team_formation = _get_team_formation(team_formation)
        home_team_formation = _get_team_formation(opponent_formation)
    else:
        raise DeserializationError(f"Unknown team_id {team.team_id}")

    return home_team_formation, away_team_formation


def identify_synthetic_formation_change_event(
    raw_event, raw_next_event, teams, home_team, away_team
):
    current_event_team = teams[str(raw_event["team"]["id"])]
    next_event_team = teams[str(raw_next_event["team"]["id"])]
    event_formation_change_info = {}
    (
        current_home_team_formation,
        current_away_team_formation,
    ) = get_home_away_team_formation(raw_event, current_event_team)
    (
        next_home_team_formation,
        next_away_team_formation,
    ) = get_home_away_team_formation(raw_next_event, next_event_team)
    if next_home_team_formation != current_home_team_formation:
        event_formation_change_info[home_team] = {
            "formation_type": next_home_team_formation
        }

    if next_away_team_formation != current_away_team_formation:
        event_formation_change_info[away_team] = {
            "formation_type": next_away_team_formation
        }

    return event_formation_change_info


def _players_to_dict(players: List[Player]):
    return {player.player_id: player for player in players}


def _parse_period_id(raw_period: str) -> int:
    if "H" in raw_period:
        period_id = int(raw_period.replace("H", ""))
    elif "E" in raw_period:
        period_id = 2 + int(raw_period.replace("E", ""))
    elif raw_period == "P":
        period_id = 5
    else:
        raise DeserializationError(f"Unknown period {raw_period}")

    return period_id


def create_periods(raw_events, period_minutes_offset_mapping):
    periods = []

    for idx, raw_event in enumerate(raw_events["events"]):
        next_period_id = None
        if (idx + 1) < len(raw_events["events"]):
            next_event = raw_events["events"][idx + 1]
            next_period_id = _parse_period_id(next_event["matchPeriod"])

        period_id = _parse_period_id(raw_event["matchPeriod"])

        if len(periods) == 0 or periods[-1].id != period_id:
            periods.append(
                Period(
                    id=period_id,
                    start_timestamp=(
                        timedelta(seconds=0)
                        if len(periods) == 0
                        else periods[-1].end_timestamp
                    ),
                    end_timestamp=None,
                )
            )

        if next_period_id != period_id:
            period_start_timestamp = periods[period_id - 1].start_timestamp
            period_offset = (
                period_start_timestamp
                - period_minutes_offset_mapping[period_id]
            )
            periods[-1] = replace(
                periods[-1],
                end_timestamp=period_offset
                + timedelta(
                    seconds=float(
                        raw_event["second"] + raw_event["minute"] * 60
                    )
                ),
            )
    return periods


class WyscoutDeserializerV3(EventDataDeserializer[WyscoutInputs]):
    @property
    def provider(self) -> Provider:
        return Provider.WYSCOUT

    def deserialize(self, inputs: WyscoutInputs) -> EventDataset:
        transformer = self.get_transformer()

        with performance_logging("load data", logger=logger):
            raw_events = json.load(inputs.event_data)
            for event in raw_events["events"]:
                if "id" not in event:
                    event["id"] = event["type"]["primary"]

        # start timestamps are fixed
        start_ts = {
            1: timedelta(minutes=0),
            2: timedelta(minutes=45),
            3: timedelta(minutes=90),
            4: timedelta(minutes=105),
            5: timedelta(minutes=120),
        }

        with performance_logging("parse data", logger=logger):
            home_team_id, away_team_id = raw_events["teams"].keys()
            home_team = _parse_team(raw_events, home_team_id, Ground.HOME)
            away_team = _parse_team(raw_events, away_team_id, Ground.AWAY)
            teams = {home_team_id: home_team, away_team_id: away_team}
            players = {
                wyId: _players_to_dict(team.players)
                for wyId, team in teams.items()
            }
            date = raw_events["match"].get("dateutc")
            if date:
                date = datetime.strptime(date, "%Y-%m-%d %H:%M:%S").replace(
                    tzinfo=timezone.utc
                )
            game_week = raw_events["match"].get("gameweek")
            if game_week:
                game_week = str(game_week)
            game_id = raw_events["events"][0].get("matchId")
            if game_id:
                game_id = str(game_id)
            home_coach = None
            away_coach = None
            coaches = raw_events.get("coaches")
            if coaches:
                if (
                    home_team_id in coaches
                    and "coach" in coaches[home_team_id]
                ):
                    home_coach = coaches[home_team_id]["coach"].get(
                        "shortName"
                    )
                if (
                    away_team_id in coaches
                    and "coach" in coaches[away_team_id]
                ):
                    away_coach = coaches[away_team_id]["coach"].get(
                        "shortName"
                    )

            periods = create_periods(raw_events, start_ts)

            events = []

            next_pass_is_kickoff = False
            for idx, raw_event in enumerate(raw_events["events"]):
                next_event = None
                ball_owning_team = None
                if (idx + 1) < len(raw_events["events"]):
                    next_event = raw_events["events"][idx + 1]

                if raw_event["possession"]:
                    ball_owning_team = teams[
                        str(raw_event["possession"]["team"]["id"])
                    ]
                if (
                    idx == 0
                    or raw_event["matchPeriod"]
                    != raw_events["events"][idx - 1]["matchPeriod"]
                    or "conceded_goal" in raw_event["type"]["secondary"]
                ):
                    next_pass_is_kickoff = True

                team_id = str(raw_event["team"]["id"])
                team = teams[team_id]
                player_id = str(raw_event["player"]["id"])
                period_id = _parse_period_id(raw_event["matchPeriod"])

                if player_id == INVALID_PLAYER:
                    player = None
                elif player_id not in players[team_id]:
                    player = None
                    warnings.warn(
                        f"Event {raw_event['id']} was performed by player {player_id} and team {team_id}, "
                        f"but the player does not appear to be part of that team's lineup. "
                        f"Handled by setting the event's player to None.",
                        DeserializationWarning,
                    )
                else:
                    player = players[team_id][player_id]

                generic_event_args = {
                    "event_id": raw_event["id"],
                    "raw_event": raw_event,
                    "coordinates": (
                        Point(
                            x=float(raw_event["location"]["x"]),
                            y=float(raw_event["location"]["y"]),
                        )
                        if raw_event["location"]
                        else None
                    ),
                    "team": team,
                    "player": player,
                    "ball_owning_team": ball_owning_team,
                    "ball_state": None,
                    "period": periods[period_id - 1],
                    "timestamp": _create_timestamp_timedelta(
                        raw_event, start_ts, period_id
                    ),
                }

                primary_event_type = raw_event["type"]["primary"]
                secondary_event_types = raw_event["type"]["secondary"]
                if (
                    primary_event_type == "shot"
                    or primary_event_type == "own_goal"
                ):
                    shot_event_args = _parse_shot(raw_event)
                    event = self.event_factory.build_shot(
                        **shot_event_args, **generic_event_args
                    )
                elif primary_event_type == "pass":
                    pass_event_args = _parse_pass(raw_event, next_event, team)
                    # Pass in new period or after goal scored is the kick-off
                    if next_pass_is_kickoff:
                        pass_event_args["qualifiers"].append(
                            SetPieceQualifier(value=SetPieceType.KICK_OFF)
                        )
                        next_pass_is_kickoff = False
                    event = self.event_factory.build_pass(
                        **pass_event_args, **generic_event_args
                    )
                elif primary_event_type == "duel":
                    if "dribble" in secondary_event_types:
                        takeon_event_args = _parse_take_on(raw_event)
                        event = self.event_factory.build_take_on(
                            **takeon_event_args, **generic_event_args
                        )
                    else:
                        duel_event_args = _parse_duel(raw_event)
                        event = self.event_factory.build_duel(
                            **duel_event_args, **generic_event_args
                        )
                elif primary_event_type == "clearance":
                    clearance_event_args = _parse_clearance(raw_event)
                    event = self.event_factory.build_clearance(
                        **clearance_event_args, **generic_event_args
                    )
                elif primary_event_type == "interception":
                    interception_event_args = _parse_interception(
                        raw_event, next_event
                    )
                    event = self.event_factory.build_interception(
                        **interception_event_args, **generic_event_args
                    )
                elif (primary_event_type == "shot_against") & (
                    "save" in raw_event["type"]["secondary"]
                ):
                    goalkeeper_save_args = _parse_goalkeeper_save(raw_event)
                    event = self.event_factory.build_goalkeeper_event(
                        **goalkeeper_save_args, **generic_event_args
                    )
                elif (
                    (primary_event_type in ["throw_in", "goal_kick"])
                    or (
                        primary_event_type == "free_kick"
                        and "free_kick_shot" not in secondary_event_types
                    )
                    or (
                        primary_event_type == "corner"
                        and "shot" not in secondary_event_types
                    )
                ):
                    set_piece_event_args = _parse_set_piece(
                        raw_event, next_event, team
                    )
                    event = self.event_factory.build_pass(
                        **set_piece_event_args, **generic_event_args
                    )
                elif (
                    (primary_event_type == "penalty")
                    or (
                        primary_event_type == "free_kick"
                        and "free_kick_shot" in secondary_event_types
                    )
                    or (
                        primary_event_type == "corner"
                        and "shot" in secondary_event_types
                    )
                ):
                    set_piece_event_args = _parse_set_piece(
                        raw_event, next_event, team
                    )
                    event = self.event_factory.build_shot(
                        **set_piece_event_args, **generic_event_args
                    )
                elif primary_event_type == "infraction":
                    if "foul" in secondary_event_types:
                        foul_event_args = _parse_foul(raw_event)
                        event = self.event_factory.build_foul_committed(
                            **foul_event_args, **generic_event_args
                        )
                        # We already append event to events
                        # as we potentially have a card and foul event for one raw event
                        if event and self.should_include_event(event):
                            events.append(transformer.transform_event(event))
                        continue
                    if (
                        "yellow_card" in secondary_event_types
                        or "red_card" in secondary_event_types
                    ):
                        card_event_args = _parse_card(raw_event)
                        event = self.event_factory.build_card(
                            **card_event_args, **generic_event_args
                        )
                        if event and self.should_include_event(event):
                            events.append(transformer.transform_event(event))
                        continue
                elif "carry" in secondary_event_types:
                    carry_event_args = _parse_carry(
                        raw_event, next_event, start_ts
                    )
                    event = self.event_factory.build_carry(
                        **carry_event_args, **generic_event_args
                    )

                else:
                    event = self.event_factory.build_generic(
                        result=None,
                        qualifiers=_generic_qualifiers(raw_event),
                        event_name=raw_event["type"]["primary"],
                        **generic_event_args,
                    )

                if event and self.should_include_event(event):
                    events.append(transformer.transform_event(event))

                if next_event:
                    event_formation_change_info = (
                        identify_synthetic_formation_change_event(
                            raw_event,
                            next_event,
                            teams,
                            home_team,
                            away_team,
                        )
                    )
                    for (
                        formation_change_team,
                        formation_change_event_kwargs,
                    ) in event_formation_change_info.items():
                        generic_event_args.update(
                            {
                                "event_id": f"synthetic-{formation_change_team.team_id}-{raw_event['id']}",
                                "raw_event": None,
                                "coordinates": None,
                                "player": None,
                                "team": formation_change_team,
                            }
                        )
                        event = self.event_factory.build_formation_change(
                            result=None,
                            qualifiers=None,
                            **formation_change_event_kwargs,
                            **generic_event_args,
                        )
                        if event and self.should_include_event(event):
                            events.append(transformer.transform_event(event))

        metadata = Metadata(
            teams=[home_team, away_team],
            periods=periods,
            pitch_dimensions=transformer.get_to_coordinate_system().pitch_dimensions,
            score=None,
            frame_rate=None,
            orientation=Orientation.ACTION_EXECUTING_TEAM,
            flags=None,
            provider=Provider.WYSCOUT,
            coordinate_system=transformer.get_to_coordinate_system(),
            date=date,
            game_week=game_week,
            game_id=game_id,
            home_coach=home_coach,
            away_coach=away_coach,
        )

        return EventDataset(metadata=metadata, records=events)



from kloppy.domain import Optional, TrackingDataset
from kloppy.infra.serializers.tracking.pff import (
    PFF_TrackingDeserializer,
    PFF_TrackingInputs,
)
from kloppy.io import FileLike, open_as_file


def load_tracking(
    meta_data: FileLike,
    roster_meta_data: FileLike,
    raw_data: FileLike,
    sample_rate: Optional[float] = None,
    limit: Optional[int] = None,
    coordinates: Optional[str] = None,
    only_alive: Optional[bool] = True,
) -> TrackingDataset:
    """
    Load and deserialize tracking data from the provided metadata, roster metadata, and raw data files.

    Args:
        meta_data (FileLike): A file-like object containing metadata about the tracking data.
        roster_meta_data (FileLike): A file-like object containing roster metadata, such as player details.
        raw_data (FileLike): A file-like object containing the raw tracking data.
        sample_rate (Optional[float], optional): The sampling rate to downsample the data. If None, no downsampling is applied. Defaults to None.
        limit (Optional[int], optional): The maximum number of records to process. If None, all records are processed. Defaults to None.
        coordinates (Optional[str], optional): The coordinate system to use for the tracking data (e.g., "pff"). Defaults to None.
        only_alive (Optional[bool], optional): Whether to include only sequences when the ball is in play. Defaults to True.

    Returns:
        TrackingDataset: A deserialized TrackingDataset object containing the processed tracking data.
    """
    deserializer = PFF_TrackingDeserializer(
        sample_rate=sample_rate,
        limit=limit,
        coordinate_system=coordinates,
        only_alive=only_alive,
    )
    with open_as_file(meta_data) as meta_data_fp, open_as_file(
        roster_meta_data
    ) as roster_meta_data_fp, open_as_file(raw_data) as raw_data_fp:
        return deserializer.deserialize(
            inputs=PFF_TrackingInputs(
                meta_data=meta_data_fp,
                roster_meta_data=roster_meta_data_fp,
                raw_data=raw_data_fp,
            )
        )



import math
from datetime import datetime, timedelta, timezone

import pytest

from kloppy import opta
from kloppy.domain import (
    BallState,
    BodyPart,
    BodyPartQualifier,
    CardEvent,
    CardType,
    CounterAttackQualifier,
    DatasetFlag,
    DatasetType,
    DuelQualifier,
    DuelType,
    EventDataset,
    EventType,
    FormationType,
    GoalkeeperActionType,
    GoalkeeperQualifier,
    OptaPitchDimensions,
    Orientation,
    PassQualifier,
    PassResult,
    PassType,
    Point,
    Point3D,
    PositionType,
    Provider,
    Score,
    SetPieceQualifier,
    SetPieceType,
    ShotResult,
    build_coordinate_system,
)
from kloppy.infra.serializers.event.statsperform.deserializer import (
    _get_end_coordinates,
)
from kloppy.infra.serializers.event.statsperform.parsers.f24_xml import (
    _parse_f24_datetime,
)


@pytest.fixture(scope="module")
def dataset(base_dir) -> EventDataset:
    """Load Opta data for FC København - FC Nordsjælland"""
    dataset = opta.load(
        f7_data=base_dir / "files" / "opta_f7.xml",
        f24_data=base_dir / "files" / "opta_f24.xml",
        coordinates="opta",
    )
    assert dataset.dataset_type == DatasetType.EVENT
    return dataset


@pytest.fixture(scope="module")
def dataset_f73(base_dir) -> EventDataset:
    """Load Opta data for FC København - FC Nordsjælland"""
    dataset = opta.load(
        f7_data=base_dir / "files" / "opta_f7.xml",
        f24_data=base_dir / "files" / "opta_f73.xml",
        coordinates="opta",
    )
    assert dataset.dataset_type == DatasetType.EVENT
    return dataset


def test_parse_f24_datetime():
    """Test if the F24 datetime is correctly parsed"""
    # timestamps have millisecond precision
    assert _parse_f24_datetime("2018-09-23T15:02:13.608") == datetime(
        2018, 9, 23, 14, 2, 13, 608000, tzinfo=timezone.utc
    )
    # milliseconds are not left-padded
    assert _parse_f24_datetime("2018-09-23T15:02:14.39") == datetime(
        2018, 9, 23, 14, 2, 14, 39000, tzinfo=timezone.utc
    )


class TestOptaMetadata:
    """Tests related to deserializing metadata (i.e., the F7 feed)"""

    def test_provider(self, dataset):
        """It should set the Opta provider"""
        assert dataset.metadata.provider == Provider.OPTA

    def test_orientation(self, dataset):
        """It should set the action-executing-team orientation"""
        assert (
            dataset.metadata.orientation == Orientation.ACTION_EXECUTING_TEAM
        )

    def test_framerate(self, dataset):
        """It should set the frame rate to None"""
        assert dataset.metadata.frame_rate is None

    def test_teams(self, dataset):
        """It should create the teams and player objects"""
        # There should be two teams with the correct names and starting formations
        assert dataset.metadata.teams[0].name == "FC København"
        assert dataset.metadata.teams[0].starting_formation == FormationType(
            "4-4-2"
        )
        assert dataset.metadata.teams[1].name == "FC Nordsjælland"
        assert dataset.metadata.teams[1].starting_formation == FormationType(
            "4-3-3"
        )
        # The teams should have the correct players
        player = dataset.metadata.teams[0].get_player_by_id("111319")
        assert player.player_id == "111319"
        assert player.jersey_no == 21
        assert str(player) == "Jesse Joronen"

    def test_player_position(self, dataset):
        """It should set the correct player position from the events"""
        # Starting players have a position
        player = dataset.metadata.teams[0].get_player_by_id("111319")
        assert player.positions.last() == PositionType.Goalkeeper
        assert player.starting

        # Substituted players don't have a position
        sub_player = dataset.metadata.teams[0].get_player_by_id("88022")
        assert len(sub_player.positions.items) == 0
        assert not sub_player.starting

        # LB position is correctly based on Formation_Place
        player = dataset.metadata.teams[0].get_player_by_id("80398")
        assert player.positions.last() == PositionType.LeftBack
        assert player.starting

    def test_periods(self, dataset):
        """It should create the periods"""
        assert len(dataset.metadata.periods) == 5
        assert dataset.metadata.periods[0].id == 1
        period_starts = [
            _parse_f24_datetime("2018-09-23T15:02:13.608"),
            _parse_f24_datetime("2018-09-23T16:05:28.873"),
            _parse_f24_datetime("2018-09-23T17:50:01.810"),
            _parse_f24_datetime("2018-09-23T18:35:01.810"),
            _parse_f24_datetime("2018-09-23T19:05:01.810"),
        ]
        period_ends = [
            _parse_f24_datetime("2018-09-23T15:48:21.222"),
            _parse_f24_datetime("2018-09-23T16:55:37.788"),
            _parse_f24_datetime("2018-09-23T18:20:01.810"),
            _parse_f24_datetime("2018-09-23T18:50:01.810"),
            _parse_f24_datetime("2018-09-23T19:25:01.810"),
        ]
        for i, period in enumerate(dataset.metadata.periods):
            assert period.id == i + 1
            assert period.start_timestamp == period_starts[i]
            assert period.end_timestamp == period_ends[i]

    def test_pitch_dimensions(self, dataset):
        """It should set the correct pitch dimensions"""
        assert dataset.metadata.pitch_dimensions == OptaPitchDimensions()

    def test_coordinate_system(self, dataset):
        """It should set the correct coordinate system"""
        assert dataset.metadata.coordinate_system == build_coordinate_system(
            Provider.OPTA
        )

    def test_score(self, dataset):
        """It should set the correct score"""
        assert dataset.metadata.score == Score(home=2, away=1)

    def test_flags(self, dataset):
        """It should set the correct flags"""
        assert (
            dataset.metadata.flags
            == DatasetFlag.BALL_OWNING_TEAM | DatasetFlag.BALL_STATE
        )


class TestOptaEvent:
    """Generic tests related to deserializing events (i.e., the F24 feed)"""

    def test_generic_attributes(self, dataset: EventDataset):
        """Test generic event attributes"""
        event = dataset.get_event_by_id("1510681159")
        assert event.event_id == "1510681159"
        assert event.team.name == "FC København"
        assert event.ball_owning_team.name == "FC København"
        assert event.player.full_name == "Dame N'Doye"
        assert event.coordinates == Point(50.1, 49.4)
        assert event.raw_event.id == "1510681159"
        assert event.related_event_ids == []
        assert event.period.id == 1
        assert event.timestamp == (
            _parse_f24_datetime("2018-09-23T15:02:14.39")  # event timestamp
            - _parse_f24_datetime("2018-09-23T15:02:13.608")  # period start
        )
        assert event.ball_state == BallState.ALIVE

    def test_timestamp(self, dataset):
        """It should set the correct timestamp, reset to zero after each period"""
        kickoff_p1 = dataset.get_event_by_id("1510681159")
        assert kickoff_p1.timestamp == timedelta(seconds=0.431)
        kickoff_p2 = dataset.get_event_by_id("1209571018")
        assert kickoff_p2.timestamp == timedelta(seconds=1.557)

    def test_correct_normalized_deserialization(self, base_dir):
        """Test if the normalized deserialization is correct"""
        dataset = opta.load(
            f7_data=base_dir / "files" / "opta_f7.xml",
            f24_data=base_dir / "files" / "opta_f24.xml",
        )
        event = dataset.get_event_by_id("1510681159")
        assert event.coordinates.x == pytest.approx(0.501, 0.001)
        assert event.coordinates.y == pytest.approx(0.50672, 0.001)

    def test_ball_owning_team(self, dataset: EventDataset):
        """Test if the ball owning team is correctly set"""
        assert (
            dataset.get_event_by_id("1594254267").ball_owning_team
            == dataset.metadata.teams[1]
        )
        assert (
            dataset.get_event_by_id("2087733359").ball_owning_team
            == dataset.metadata.teams[0]
        )

    def test_setpiece_qualifiers(self, dataset: EventDataset):
        """Test if the qualifiers are correctly deserialized"""
        kick_off = dataset.get_event_by_id("1510681159")
        assert (
            kick_off.get_qualifier_value(SetPieceQualifier)
            == SetPieceType.KICK_OFF
        )

    def test_body_part_qualifiers(self, dataset: EventDataset):
        """Test if the body part qualifiers are correctly deserialized"""
        header = dataset.get_event_by_id("1101592119")
        assert BodyPart.HEAD in header.get_qualifier_values(BodyPartQualifier)

    def test_card_event(self, dataset: EventDataset):
        """Test if the card qualifiers are correctly deserialized"""
        red_card = dataset.get_event_by_id("2318454729")
        assert isinstance(red_card, CardEvent)
        assert red_card.card_type == CardType.RED

    def test_counter_attack_qualifiers(self, dataset: EventDataset):
        """Test if the counter attack qualifiers are correctly deserialized"""
        counter_attack = dataset.get_event_by_id("2318695229")
        assert (
            counter_attack.get_qualifier_value(CounterAttackQualifier) is True
        )


class TestOptaPassEvent:
    """Tests related to deserializing pass events"""

    def test_deserialize_all(self, dataset: EventDataset):
        """It should deserialize all pass events"""
        events = dataset.find_all("pass")
        assert len(events) == 16

    def test_receiver_coordinates(self, dataset: EventDataset):
        """Test if the receiver coordinates are correctly deserialized"""
        # Check receiver coordinates for incomplete passes
        incomplete_pass = dataset.get_event_by_id("1101592119")
        assert incomplete_pass.receiver_coordinates.x == 45.5
        assert incomplete_pass.receiver_coordinates.y == 68.2

    def test_end_coordinates(self, dataset: EventDataset):
        """Test if the end coordinates are correctly deserialized"""
        pass_event = dataset.get_event_by_id("2360555167")
        assert pass_event.receiver_coordinates.x == 89.3

    def test_pass_qualifiers(self, dataset: EventDataset):
        """Test if the pass type qualfiers are correctly deserialized"""
        through_ball = dataset.get_event_by_id("1101592119")
        assert PassType.THROUGH_BALL in through_ball.get_qualifier_values(
            PassQualifier
        )
        chipped_pass = dataset.get_event_by_id("1444075194")
        assert PassType.CHIPPED_PASS in chipped_pass.get_qualifier_values(
            PassQualifier
        )
        assist = dataset.get_event_by_id("1666666666")
        assert PassType.ASSIST in assist.get_qualifier_values(PassQualifier)
        assert PassType.SHOT_ASSIST in assist.get_qualifier_values(
            PassQualifier
        )

    def test_pass_qualifiers_for_deflected_pass(self, dataset: EventDataset):
        """Test if the pass type qualfiers are correctly deserialized for deflected passes"""
        deflected_pass = dataset.get_event_by_id("2509132176")
        assert deflected_pass.result == PassResult.COMPLETE
        assert deflected_pass.receiver_coordinates.x == 3.3
        assert deflected_pass.receiver_coordinates.y == 81.1

    def test_ball_state(self, dataset: EventDataset):
        """Test if the ball state is correctly set"""
        events = dataset.find_all("pass")
        assert all(
            event.ball_state == BallState.ALIVE for event in events
        ), "Not all pass ball states are ALIVE"


class TestOptaClearanceEvent:
    """Tests related to deserializing clearance events"""

    def test_deserialize_all(self, dataset: EventDataset):
        """It should deserialize all clearance events"""
        events = dataset.find_all("clearance")
        assert len(events) == 2

    def test_correct_deserialization(self, dataset: EventDataset):
        """Test if the clearance event is correctly deserialized"""
        clearance = dataset.get_event_by_id("2498907287")
        assert clearance.event_type == EventType.CLEARANCE


class TestOptaShotEvent:
    """Tests related to deserializing shot events"""

    def test_deserialize_all(self, dataset: EventDataset):
        """It should deserialize all shot events"""
        events = dataset.find_all("shot")
        assert len(events) == 4

    def test_correct_deserialization(self, dataset: EventDataset):
        """Test if the shot event is correctly deserialized"""
        shot = dataset.get_event_by_id("2318695229")
        # A shot event should have a result
        assert shot.result == ShotResult.GOAL
        # A shot event should have end coordinates
        assert shot.result_coordinates == Point3D(100.0, 47.8, 2.5)
        # A shot event should have a body part
        assert (
            shot.get_qualifier_value(BodyPartQualifier) == BodyPart.LEFT_FOOT
        )

    def test_timestamp_goal(self, dataset: EventDataset):
        """Check timestamp from qualifier 374 in case of goal"""
        goal = dataset.get_event_by_id("2318695229")
        assert goal.timestamp == (
            _parse_f24_datetime("2018-09-23T16:07:48.525")  # event timestamp
            - _parse_f24_datetime("2018-09-23T16:05:28.873")  # period start
        )

    def test_shot_end_coordinates(self):
        """Shots should receive the correct end coordinates."""
        # When no end coordinates are available, we return None
        assert _get_end_coordinates({}) is None

        # When a shot was not blocked, the goalmouth coordinates should be used.
        # The y- and z-coordinate are specified by qualifiers; the
        # x-coordinate is 100.0 (i.e., the goal line)
        shot_on_target_qualifiers = {
            102: "52.1",  # goal mouth y-coordinate
            103: "18.4",  # goal mouth z-coordinate
        }
        assert _get_end_coordinates(shot_on_target_qualifiers) == Point3D(
            x=100.0, y=52.1, z=18.4
        )

        # When the z-coordinate is missing, we return 2D coordinates
        incomplete_shot_qualifiers = {
            102: "52.1",  # goal mouth y-coordinate
        }
        assert _get_end_coordinates(incomplete_shot_qualifiers) == Point(
            x=100, y=52.1
        )

        # When the y-coordinate is missing, we return None
        incomplete_shot_qualifiers = {
            103: "18.4",  # goal mouth z-coordinate
        }
        assert _get_end_coordinates(incomplete_shot_qualifiers) is None

        # When a shot is blocked, the end coordinates should correspond to the
        # location where the shot was blocked.
        blocked_shot_qualifiers = {
            146: "99.1",  # blocked x-coordiante
            147: "52.5",  # blocked y-coordinate
        }
        assert _get_end_coordinates(blocked_shot_qualifiers) == Point(
            x=99.1, y=52.5
        )

        # When a shot was blocked and goal mouth locations are provided too,
        # the z-coordinate of the goal mouth coordinates should be inversely
        # projected on the location where the shot was blocked
        blocked_shot_on_target_qualifiers = {
            **shot_on_target_qualifiers,
            **blocked_shot_qualifiers,
        }
        start_coordinates = Point(x=92.6, y=57.9)
        # This requires some trigonometry. We can define two
        # right-angle triangles:
        #   - a large triangle between the start coordinates and goal mouth
        #     coordinates.
        #   - an enclosed smaller triangle between the start coordinates and
        #     the location where the shot was blocked.
        # We need to compute the length of the opposite side of the small
        # triangle. Therefore, we compute:
        #   - the length of the adjacent side of the large triangle
        adj_large = math.sqrt(
            (100 - start_coordinates.x) ** 2
            + (52.1 - start_coordinates.y) ** 2
        )
        #   - the length of the adjacent side of the small triangle
        adj_small = math.sqrt(
            (99.1 - start_coordinates.x) ** 2
            + (52.5 - start_coordinates.y) ** 2
        )
        #   - the angle of the large triangle (== the angle of the small triangle)
        alpha_large = math.atan2(18.4, adj_large)
        #  - the opposite side of the small triangle
        opp_small = math.tan(alpha_large) * adj_small
        assert _get_end_coordinates(
            blocked_shot_on_target_qualifiers, start_coordinates
        ) == Point3D(x=99.1, y=52.5, z=opp_small)

    def test_own_goal(self, dataset: EventDataset):
        """Test if own goals are correctly deserialized"""
        own_goal = dataset.get_event_by_id("2318697001")
        assert own_goal.result == ShotResult.OWN_GOAL
        # Use the inverse coordinates of the goal location
        assert own_goal.result_coordinates == Point3D(0.0, 100 - 45.6, 1.9)
        assert own_goal.ball_state == BallState.DEAD

    def test_goal(self, dataset: EventDataset):
        """Test if goals are correctly deserialized"""
        goal = dataset.get_event_by_id("2614247749")
        assert goal.result == ShotResult.GOAL
        assert (
            next(
                statistic
                for statistic in goal.statistics
                if statistic.name == "xG"
            ).value
            == 0.9780699610710144
        )
        assert (
            next(
                statistic
                for statistic in goal.statistics
                if statistic.name == "PSxG"
            ).value
            == 0.98
        )

    def test_shot_xg(self, dataset_f73: EventDataset):
        """Test if expected goals are correctly deserialized"""
        shot = dataset_f73.get_event_by_id("2318695229")
        assert (
            next(
                statistic
                for statistic in shot.statistics
                if statistic.name == "xG"
            ).value
            == 0.75
        )


class TestOptaDuelEvent:
    """Tests related to deserializing duel events"""

    def test_deserialize_all(self, dataset: EventDataset):
        """It should deserialize all duel events"""
        events = dataset.find_all("duel")
        assert len(events) == 3

    def test_qualifiers(self, dataset: EventDataset):
        """Test if the qualifiers are correctly deserialized"""
        aerial_duel = dataset.get_event_by_id("1274474573")
        assert DuelType.AERIAL in aerial_duel.get_qualifier_values(
            DuelQualifier
        )
        ground_duel = dataset.get_event_by_id("2140914735")
        assert DuelType.GROUND in ground_duel.get_qualifier_values(
            DuelQualifier
        )


class TestOptaGoalkeeperEvent:
    """Tests related to deserializing goalkeeper events"""

    def test_deserialize_all(self, dataset: EventDataset):
        """It should deserialize all goalkeeper events"""
        events = dataset.find_all("goalkeeper")
        assert len(events) == 5

    def test_qualifiers(self, dataset: EventDataset):
        """Test if the qualifiers are correctly deserialized"""
        save = dataset.get_event_by_id("2451170467")
        assert (
            save.get_qualifier_value(GoalkeeperQualifier)
            == GoalkeeperActionType.SAVE
        )
        claim = dataset.get_event_by_id("2453149143")
        assert (
            claim.get_qualifier_value(GoalkeeperQualifier)
            == GoalkeeperActionType.CLAIM
        )
        punch = dataset.get_event_by_id("2451094707")
        assert (
            punch.get_qualifier_value(GoalkeeperQualifier)
            == GoalkeeperActionType.PUNCH
        )
        keeper_pick_up = dataset.get_event_by_id("2451098837")
        assert (
            keeper_pick_up.get_qualifier_value(GoalkeeperQualifier)
            == GoalkeeperActionType.PICK_UP
        )
        smother = dataset.get_event_by_id("2438594253")
        assert (
            smother.get_qualifier_value(GoalkeeperQualifier)
            == GoalkeeperActionType.SMOTHER
        )


class TestOptaInterceptionEvent:
    """Tests related to deserializing interception events"""

    def test_correct_deserialization(self, dataset: EventDataset):
        """Test if the interception event is correctly deserialized"""
        event = dataset.get_event_by_id("2609934569")
        assert event.event_type == EventType.INTERCEPTION


class TestOptaMiscontrolEvent:
    """Tests related to deserializing miscontrol events"""

    def test_correct_deserialization(self, dataset: EventDataset):
        """Test if the miscontrol event is correctly deserialized"""
        event = dataset.get_event_by_id("2509132175")
        assert event.event_type == EventType.MISCONTROL


class TestOptaBlockEvent:
    """Tests related to deserializing block events"""

    def test_correct_deserialization(self, dataset: EventDataset):
        """Test if the block event is correctly deserialized"""
        event = dataset.get_event_by_id("1515097981")
        assert event.event_type == EventType.GENERIC
        assert event.event_name == "block"


class TestOptaCardEvent:
    """Tests related to deserialzing card events"""

    def test_deserialize_all(self, dataset: EventDataset):
        """It should deserialize all card events"""
        events = dataset.find_all("card")
        assert len(events) == 1
        assert all(
            event.ball_state == BallState.DEAD for event in events
        ), "Not all card ball states are ALIVE"



import math
import sys
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional, Set, Type, Union

from kloppy.domain import (
    BodyPartQualifier,
    Code,
    Event,
    Frame,
    Orientation,
    Point,
    QualifierMixin,
    ResultMixin,
)
from kloppy.domain.models.event import (
    CardEvent,
    CarryEvent,
    EnumQualifier,
    EventType,
    PassEvent,
    ShotEvent,
)
from kloppy.exceptions import (
    KloppyParameterError,
    OrientationError,
    UnknownEncoderError,
)
from kloppy.utils import camelcase_to_snakecase

py_version = sys.version_info
if py_version >= (3, 8):
    from typing import get_args


def _get_generic_type_arg(cls):
    t = cls.__orig_bases__[0]
    if py_version >= (3, 8):
        return get_args(t)[0]
    else:
        return t.__args__[0]


class OneHotEncoder:
    def __init__(self, name: str, options: List[str]):
        self.options = options
        self.name = name

    def encode(self, value: Union[Set, str]) -> Dict[str, bool]:
        if isinstance(value, str):
            value = [value]

        return {
            f"is_{self.name}_{option_.lower()}": option_ in value
            for option_ in self.options
        }


class EventAttributeTransformer(ABC):
    @abstractmethod
    def __call__(self, event: Event) -> Dict[str, Any]:
        pass


class AngleToGoalTransformer(EventAttributeTransformer):
    def __call__(self, event: Event) -> Dict[str, Any]:
        metadata = event.dataset.metadata
        if metadata.orientation != Orientation.ACTION_EXECUTING_TEAM:
            raise OrientationError(
                "Can only calculate Angle when dataset orientation is ACTION_EXECUTING_TEAM"
            )

        if not event.coordinates:
            return {"angle_to_goal": None}
        delta_x = metadata.pitch_dimensions.distance_between(
            Point(event.coordinates.x, 0),
            Point(metadata.pitch_dimensions.x_dim.max, 0),
        )
        delta_y = metadata.pitch_dimensions.distance_between(
            Point(0, event.coordinates.y),
            Point(
                0,
                (
                    metadata.pitch_dimensions.y_dim.max
                    + metadata.pitch_dimensions.y_dim.min
                )
                / 2,
            ),
        )

        return {"angle_to_goal": math.atan2(delta_x, delta_y) / math.pi * 180}


class DistanceToGoalTransformer(EventAttributeTransformer):
    def __call__(self, event: Event) -> Dict[str, Any]:
        metadata = event.dataset.metadata
        if not event.coordinates:
            return {"distance_to_goal": None}

        event_x = event.coordinates.x
        event_y = event.coordinates.y
        goal_x = metadata.pitch_dimensions.x_dim.max
        goal_y = (
            metadata.pitch_dimensions.y_dim.max
            + metadata.pitch_dimensions.y_dim.min
        ) / 2

        return {
            "distance_to_goal": math.sqrt(
                (goal_x - event_x) ** 2 + (goal_y - event_y) ** 2
            )
        }


class DistanceToOwnGoalTransformer(EventAttributeTransformer):
    def __call__(self, event: Event) -> Dict[str, Any]:
        metadata = event.dataset.metadata

        if not event.coordinates:
            return {"distance_to_own_goal": None}

        event_x = event.coordinates.x
        event_y = event.coordinates.y
        goal_x = metadata.pitch_dimensions.x_dim.min
        goal_y = (
            metadata.pitch_dimensions.y_dim.max
            + metadata.pitch_dimensions.y_dim.min
        ) / 2

        return {
            "distance_to_own_goal": math.sqrt(
                (goal_x - event_x) ** 2 + (goal_y - event_y) ** 2
            )
        }


def create_transformer_from_qualifier(
    qualifier_type: Type[EnumQualifier],
) -> Type[EventAttributeTransformer]:
    enum_ = _get_generic_type_arg(qualifier_type)
    name = camelcase_to_snakecase(enum_.__name__)
    options = [e.value for e in enum_]

    class _Transformer(EventAttributeTransformer):
        def __init__(self, encoding: str = "one-hot"):
            if encoding == "one-hot":
                self.encoder = OneHotEncoder(name, options)
            else:
                raise UnknownEncoderError(f"Don't know {encoding} encoding")

        def __call__(self, event: Event) -> Dict[str, Any]:
            values = {
                qualifier.value.value
                for qualifier in event.qualifiers
                if isinstance(qualifier, qualifier_type)
            }
            return self.encoder.encode(values)

    return _Transformer


class DefaultEventTransformer(EventAttributeTransformer):
    def __init__(
        self,
        *include: str,
        exclude: Optional[List[str]] = None,
    ):
        if include and exclude:
            raise KloppyParameterError(
                "Cannot specify both include as exclude"
            )

        self.exclude = exclude or []
        self.include = include or []

    def __call__(self, event: Event) -> Dict[str, Any]:
        row = dict(
            event_id=event.event_id,
            event_type=(
                event.event_type.value
                if event.event_type != EventType.GENERIC
                else f"GENERIC:{event.event_name}"
            ),
            period_id=event.period.id,
            timestamp=event.timestamp,
            end_timestamp=None,
            ball_state=event.ball_state.value if event.ball_state else None,
            ball_owning_team=(
                event.ball_owning_team.team_id
                if event.ball_owning_team
                else None
            ),
            team_id=event.team.team_id if event.team else None,
            player_id=event.player.player_id if event.player else None,
            coordinates_x=event.coordinates.x if event.coordinates else None,
            coordinates_y=event.coordinates.y if event.coordinates else None,
        )
        if isinstance(event, PassEvent):
            row.update(
                {
                    "end_timestamp": event.receive_timestamp,
                    "end_coordinates_x": (
                        event.receiver_coordinates.x
                        if event.receiver_coordinates
                        else None
                    ),
                    "end_coordinates_y": (
                        event.receiver_coordinates.y
                        if event.receiver_coordinates
                        else None
                    ),
                    "receiver_player_id": (
                        event.receiver_player.player_id
                        if event.receiver_player
                        else None
                    ),
                }
            )
        elif isinstance(event, CarryEvent):
            row.update(
                {
                    "end_timestamp": event.end_timestamp,
                    "end_coordinates_x": (
                        event.end_coordinates.x
                        if event.end_coordinates
                        else None
                    ),
                    "end_coordinates_y": (
                        event.end_coordinates.y
                        if event.end_coordinates
                        else None
                    ),
                }
            )
        elif isinstance(event, ShotEvent):
            row.update(
                {
                    "end_coordinates_x": (
                        event.result_coordinates.x
                        if event.result_coordinates
                        else None
                    ),
                    "end_coordinates_y": (
                        event.result_coordinates.y
                        if event.result_coordinates
                        else None
                    ),
                }
            )
        elif isinstance(event, CardEvent):
            row.update(
                {
                    "card_type": (
                        event.card_type.value if event.card_type else None
                    )
                }
            )

        if isinstance(event, QualifierMixin) and event.qualifiers:
            for qualifier in event.qualifiers:
                row.update(qualifier.to_dict())

        if isinstance(event, ResultMixin) and event.result is not None:
            row.update(
                {
                    "result": event.result.value,
                    "success": event.result.is_success,
                }
            )
        else:
            row.update(
                {
                    "result": None,
                    "success": None,
                }
            )

        if self.include:
            return {k: row[k] for k in self.include}
        elif self.exclude:
            return {k: v for k, v in row.items() if k not in self.exclude}
        else:
            return row


class DefaultFrameTransformer:
    def __init__(
        self,
        *include: str,
        exclude: Optional[List[str]] = None,
    ):
        if include and exclude:
            raise KloppyParameterError(
                "Cannot specify both include as exclude"
            )

        self.exclude = exclude or []
        self.include = include or []

    def __call__(self, frame: Frame) -> Dict[str, Any]:
        row = dict(
            period_id=frame.period.id if frame.period else None,
            timestamp=frame.timestamp,
            frame_id=frame.frame_id,
            ball_state=frame.ball_state.value if frame.ball_state else None,
            ball_owning_team_id=(
                frame.ball_owning_team.team_id
                if frame.ball_owning_team
                else None
            ),
            ball_x=(
                frame.ball_coordinates.x if frame.ball_coordinates else None
            ),
            ball_y=(
                frame.ball_coordinates.y if frame.ball_coordinates else None
            ),
            ball_z=(
                getattr(frame.ball_coordinates, "z", None)
                if frame.ball_coordinates
                else None
            ),
            ball_speed=frame.ball_speed,
        )
        for player, player_data in frame.players_data.items():
            row.update(
                {
                    f"{player.player_id}_x": (
                        player_data.coordinates.x
                        if player_data.coordinates
                        else None
                    ),
                    f"{player.player_id}_y": (
                        player_data.coordinates.y
                        if player_data.coordinates
                        else None
                    ),
                    f"{player.player_id}_d": player_data.distance,
                    f"{player.player_id}_s": player_data.speed,
                }
            )

            if player_data.other_data:
                for name, value in player_data.other_data.items():
                    row.update(
                        {
                            f"{player.player_id}_{name}": value,
                        }
                    )

        if frame.other_data:
            for name, value in frame.other_data.items():
                row.update(
                    {
                        name: value,
                    }
                )

        if self.include:
            return {k: row[k] for k in self.include}
        elif self.exclude:
            return {k: v for k, v in row.items() if k not in self.exclude}
        else:
            return row


class DefaultCodeTransformer:
    def __init__(
        self,
        *include: str,
        exclude: Optional[List[str]] = None,
    ):
        if include and exclude:
            raise KloppyParameterError(
                "Cannot specify both include as exclude"
            )

        self.exclude = exclude or []
        self.include = include or []

    def __call__(self, code: Code) -> Dict[str, Any]:
        row = dict(
            code_id=code.code_id,
            period_id=code.period.id if code.period else None,
            timestamp=code.timestamp,
            end_timestamp=code.end_timestamp,
            code=code.code,
        )
        row.update(code.labels)

        if self.include:
            return {k: row[k] for k in self.include}
        elif self.exclude:
            return {k: v for k, v in row.items() if k not in self.exclude}
        else:
            return row


BodyPartTransformer = create_transformer_from_qualifier(BodyPartQualifier)



"""XML parser for Stats Perform MA3 feeds."""

from datetime import datetime, timezone
from typing import List

from .base import OptaEvent, OptaXMLParser


def _parse_ma3_datetime(dt_str: str) -> datetime:
    try:
        return datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%S.%fZ").replace(
            tzinfo=timezone.utc
        )
    except ValueError:
        return datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%SZ").replace(
            tzinfo=timezone.utc
        )


class MA3XMLParser(OptaXMLParser):
    """Extract data from a Stats Perform MA3 data stream."""

    def extract_events(self) -> List[OptaEvent]:
        live_data = self.root.liveData
        return [
            OptaEvent(
                id=event.attrib["id"],
                event_id=int(event.attrib["eventId"]),
                type_id=int(event.attrib["typeId"]),
                period_id=int(event.attrib["periodId"]),
                time_min=int(event.attrib["timeMin"]),
                time_sec=int(event.attrib["timeSec"]),
                x=float(event.attrib["x"]),
                y=float(event.attrib["y"]),
                timestamp=_parse_ma3_datetime(event.attrib["timeStamp"]),
                last_modified=_parse_ma3_datetime(
                    event.attrib["lastModified"]
                ),
                contestant_id=event.attrib.get("contestantId"),
                player_id=event.attrib.get("playerId"),
                outcome=(
                    int(event.attrib["outcome"])
                    if "outcome" in event.attrib
                    else None
                ),
                qualifiers={
                    int(qualifier.attrib["qualifierId"]): qualifier.attrib.get(
                        "value"
                    )
                    for qualifier in event.iterchildren("qualifier")
                },
            )
            for event in live_data.events.iterchildren("event")
        ]
